from datetime import datetime, timedelta
from typing import Any, Dict

from betterproto import DATETIME_ZERO, TYPE_MAP, TYPE_MESSAGE, Casing, Message


# https://github.com/danielgtaylor/python-betterproto/pull/609
def to_pydict(
    self, casing: Casing = Casing.CAMEL, include_default_values: bool = False
) -> Dict[str, Any]:
    """
    Returns a python dict representation of this object.

    Parameters
    -----------
    casing: :class:`Casing`
        The casing to use for key values. Default is :attr:`Casing.CAMEL` for
        compatibility purposes.
    include_default_values: :class:`bool`
        If ``True`` will include the default values of fields. Default is ``False``.
        E.g. an ``int32`` field will be included with a value of ``0`` if this is
        set to ``True``, otherwise this would be ignored.

    Returns
    --------
    Dict[:class:`str`, Any]
        The python dict representation of this object.
    """
    output: Dict[str, Any] = {}
    defaults = self._betterproto.default_gen
    for field_name, meta in self._betterproto.meta_by_field_name.items():
        field_is_repeated = defaults[field_name] is list
        try:
            value = getattr(self, field_name)
        except AttributeError:
            value = self._get_field_default(field_name)
        cased_name = casing(field_name).rstrip("_")  # type: ignore
        if meta.proto_type == TYPE_MESSAGE:
            if isinstance(value, datetime):
                if (
                    value != DATETIME_ZERO
                    or include_default_values
                    or self._include_default_value_for_oneof(
                        field_name=field_name, meta=meta
                    )
                ):
                    output[cased_name] = value
            elif isinstance(value, timedelta):
                if (
                    value != timedelta(0)
                    or include_default_values
                    or self._include_default_value_for_oneof(
                        field_name=field_name, meta=meta
                    )
                ):
                    output[cased_name] = value
            elif meta.wraps:
                if value is not None or include_default_values:
                    output[cased_name] = value
            elif field_is_repeated:
                # Convert each item.
                value = [i.to_pydict(casing, include_default_values) for i in value]
                if value or include_default_values:
                    output[cased_name] = value
            elif value is None:
                if include_default_values:
                    output[cased_name] = None
            elif (
                value._serialized_on_wire
                or include_default_values
                or self._include_default_value_for_oneof(
                    field_name=field_name, meta=meta
                )
            ):
                output[cased_name] = value.to_pydict(casing, include_default_values)
        elif meta.proto_type == TYPE_MAP:
            for k in value:
                if hasattr(value[k], "to_pydict"):
                    value[k] = value[k].to_pydict(casing, include_default_values)

            if value or include_default_values:
                output[cased_name] = value
        elif (
            value != self._get_field_default(field_name)
            or include_default_values
            or self._include_default_value_for_oneof(field_name=field_name, meta=meta)
        ):
            output[cased_name] = value
    return output


def patch():
    Message.to_pydict = to_pydict



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M50SPlusPlusVK10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50SPlusPlusVK10

    expected_chips = 117
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M50SPlusPlusVK20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50SPlusPlusVK20

    expected_chips = 123
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M50SPlusPlusVK30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50SPlusPlusVK30

    expected_chips = 156
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M50SPlusPlusVK40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50SPlusPlusVK40

    expected_chips = 129
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M50SPlusPlusVK50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50SPlusPlusVK50

    expected_chips = 135
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M50SPlusPlusVK60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50SPlusPlusVK60

    expected_chips = 111
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M50SPlusPlusVL20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50SPlusPlusVL20

    expected_chips = 86
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M50SPlusPlusVL30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50SPlusPlusVL30

    expected_chips = 111
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M50SPlusPlusVL40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50SPlusPlusVL40

    expected_chips = 117
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M50SPlusPlusVL50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50SPlusPlusVL50

    expected_chips = 105
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M50SPlusPlusVL60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50SPlusPlusVL60

    expected_chips = 111
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M63VK10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M63VK10

    expected_chips = None
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M63VK20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M63VK20

    expected_chips = 264
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M63VK30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M63VK30

    expected_chips = 272
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M63VL10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M63VL10

    expected_chips = 174
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M63VL30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M63VL30

    expected_chips = 216
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------

import re
from typing import List, Optional

from pyasic.data import Fan, HashBoard
from pyasic.device.algorithm import AlgoHashRate
from pyasic.errors import APIError
from pyasic.miners.backends.cgminer import CGMiner
from pyasic.miners.data import DataFunction, DataLocations, DataOptions, RPCAPICommand

AVALON_DATA_LOC = DataLocations(
    **{
        str(DataOptions.MAC): DataFunction(
            "_get_mac",
            [RPCAPICommand("rpc_version", "version")],
        ),
        str(DataOptions.API_VERSION): DataFunction(
            "_get_api_ver",
            [RPCAPICommand("rpc_version", "version")],
        ),
        str(DataOptions.FW_VERSION): DataFunction(
            "_get_fw_ver",
            [RPCAPICommand("rpc_version", "version")],
        ),
        str(DataOptions.HASHRATE): DataFunction(
            "_get_hashrate",
            [RPCAPICommand("rpc_devs", "devs")],
        ),
        str(DataOptions.EXPECTED_HASHRATE): DataFunction(
            "_get_expected_hashrate",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.HASHBOARDS): DataFunction(
            "_get_hashboards",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.ENVIRONMENT_TEMP): DataFunction(
            "_get_env_temp",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.WATTAGE_LIMIT): DataFunction(
            "_get_wattage_limit",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.WATTAGE): DataFunction(
            "_get_wattage",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.FANS): DataFunction(
            "_get_fans",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.FAULT_LIGHT): DataFunction(
            "_get_fault_light",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.UPTIME): DataFunction(
            "_get_uptime",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.POOLS): DataFunction(
            "_get_pools",
            [RPCAPICommand("rpc_pools", "pools")],
        ),
    }
)


class AvalonMiner(CGMiner):
    """Handler for Avalon Miners"""

    data_locations = AVALON_DATA_LOC

    async def fault_light_on(self) -> bool:
        try:
            data = await self.rpc.ascset(0, "led", "1-1")
        except APIError:
            return False
        if data["STATUS"][0]["Msg"] == "ASC 0 set OK":
            return True
        return False

    async def fault_light_off(self) -> bool:
        try:
            data = await self.rpc.ascset(0, "led", "1-0")
        except APIError:
            return False
        if data["STATUS"][0]["Msg"] == "ASC 0 set OK":
            return True
        return False

    async def set_power_limit(self, wattage: int) -> bool:
        try:
            if wattage < 3:
                limit = wattage
            elif wattage > 100:
                limit = 2
            elif wattage > 80:
                limit = 1
            else:
                limit = 0
            data = await self.rpc.ascset(0, "worklevel,set", 1)
        except APIError:
            return False
        if data["STATUS"][0]["Msg"] == "ASC 0 set OK":
            return True
        return False

    async def reboot(self) -> bool:
        try:
            data = await self.rpc.restart()
        except APIError:
            return False

        try:
            if data["STATUS"] == "RESTART":
                return True
        except KeyError:
            return False
        return False

    @staticmethod
    def parse_stats(stats):
        _stats_items = re.findall(".+?\\[*?]", stats)
        stats_items = []
        stats_dict = {}
        for item in _stats_items:
            if ": " in item:
                data = item.replace("]", "").split("[")
                data_list = [i.split(": ") for i in data[1].strip().split(", ")]
                data_dict = {}
                try:
                    for key, val in [tuple(item) for item in data_list]:
                        data_dict[key] = val
                except ValueError:
                    # --avalon args
                    for arg_item in data_list:
                        item_data = arg_item[0].split(" ")
                        for idx, val in enumerate(item_data):
                            if idx % 2 == 0 or idx == 0:
                                data_dict[val] = item_data[idx + 1]

                raw_data = [data[0].strip(), data_dict]
            else:
                raw_data = [
                    value
                    for value in item.replace("[", " ")
                    .replace("]", " ")
                    .split(" ")[:-1]
                    if value != ""
                ]
                if len(raw_data) == 1:
                    raw_data.append("")
            if raw_data[0] == "":
                raw_data = raw_data[1:]

            stats_dict[raw_data[0]] = raw_data[1:]
            stats_items.append(raw_data)

        return stats_dict

    ##################################################
    ### DATA GATHERING FUNCTIONS (get_{some_data}) ###
    ##################################################

    async def _get_mac(self, rpc_version: dict = None) -> Optional[str]:
        if rpc_version is None:
            try:
                rpc_version = await self.rpc.version()
            except APIError:
                pass

        if rpc_version is not None:
            try:
                base_mac = rpc_version["VERSION"][0]["MAC"]
                base_mac = base_mac.upper()
                mac = ":".join(
                    [base_mac[i : (i + 2)] for i in range(0, len(base_mac), 2)]
                )
                return mac
            except (KeyError, ValueError):
                pass

    async def _get_hashrate(self, rpc_devs: dict = None) -> Optional[AlgoHashRate]:
        if rpc_devs is None:
            try:
                rpc_devs = await self.rpc.devs()
            except APIError:
                pass

        if rpc_devs is not None:
            try:
                return self.algo.hashrate(
                    rate=float(rpc_devs["DEVS"][0]["MHS 1m"]), unit=self.algo.unit.MH
                ).into(self.algo.unit.default)
            except (KeyError, IndexError, ValueError, TypeError):
                pass

    async def _get_hashboards(self, rpc_stats: dict = None) -> List[HashBoard]:
        if self.expected_hashboards is None:
            return []

        hashboards = [
            HashBoard(slot=i, expected_chips=self.expected_chips)
            for i in range(self.expected_hashboards)
        ]

        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        if rpc_stats is not None:
            try:
                unparsed_stats = rpc_stats["STATS"][0]["MM ID0"]
                parsed_stats = self.parse_stats(unparsed_stats)
            except (IndexError, KeyError, ValueError, TypeError):
                return hashboards

            for board in range(self.expected_hashboards):
                try:
                    hashboards[board].chip_temp = int(parsed_stats["MTmax"][board])
                except LookupError:
                    pass

                try:
                    board_hr = parsed_stats["MGHS"][board]
                    hashboards[board].hashrate = self.algo.hashrate(
                        rate=float(board_hr), unit=self.algo.unit.GH
                    ).into(self.algo.unit.default)
                except LookupError:
                    pass

                try:
                    hashboards[board].temp = int(parsed_stats["MTavg"][board])
                except LookupError:
                    pass

                try:
                    chip_data = parsed_stats[f"PVT_T{board}"]
                    hashboards[board].missing = False
                    if chip_data:
                        hashboards[board].chips = len(
                            [item for item in chip_data if not item == "0"]
                        )
                except LookupError:
                    pass

        return hashboards

    async def _get_expected_hashrate(
        self, rpc_stats: dict = None
    ) -> Optional[AlgoHashRate]:
        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        if rpc_stats is not None:
            try:
                unparsed_stats = rpc_stats["STATS"][0]["MM ID0"]
                parsed_stats = self.parse_stats(unparsed_stats)
                return self.algo.hashrate(
                    rate=float(parsed_stats["GHSmm"][0]), unit=self.algo.unit.GH
                ).into(self.algo.unit.default)
            except (IndexError, KeyError, ValueError, TypeError):
                pass

    async def _get_env_temp(self, rpc_stats: dict = None) -> Optional[float]:
        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        if rpc_stats is not None:
            try:
                unparsed_stats = rpc_stats["STATS"][0]["MM ID0"]
                parsed_stats = self.parse_stats(unparsed_stats)
                return float(parsed_stats["Temp"][0])
            except (IndexError, KeyError, ValueError, TypeError):
                pass

    async def _get_wattage_limit(self, rpc_stats: dict = None) -> Optional[int]:
        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        if rpc_stats is not None:
            try:
                unparsed_stats = rpc_stats["STATS"][0]["MM ID0"]
                parsed_stats = self.parse_stats(unparsed_stats)
                return int(parsed_stats["MPO"][0])
            except (IndexError, KeyError, ValueError, TypeError):
                pass

    async def _get_wattage(self, rpc_stats: dict = None) -> Optional[int]:
        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        if rpc_stats is not None:
            try:
                unparsed_stats = rpc_stats["STATS"][0]["MM ID0"]
                parsed_stats = self.parse_stats(unparsed_stats)
                return int(parsed_stats["WALLPOWER"][0])
            except (IndexError, KeyError, ValueError, TypeError):
                pass

    async def _get_fans(self, rpc_stats: dict = None) -> List[Fan]:
        if self.expected_fans is None:
            return []

        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        fans_data = [Fan() for _ in range(self.expected_fans)]
        if rpc_stats is not None:
            try:
                unparsed_stats = rpc_stats["STATS"][0]["MM ID0"]
                parsed_stats = self.parse_stats(unparsed_stats)
            except LookupError:
                return fans_data

            for fan in range(self.expected_fans):
                try:
                    fans_data[fan].speed = int(parsed_stats[f"Fan{fan + 1}"][0])
                except (IndexError, KeyError, ValueError, TypeError):
                    pass
        return fans_data

    async def _get_fault_light(self, rpc_stats: dict = None) -> Optional[bool]:
        if self.light:
            return self.light
        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        if rpc_stats is not None:
            try:
                unparsed_stats = rpc_stats["STATS"][0]["MM ID0"]
                parsed_stats = self.parse_stats(unparsed_stats)
                led = int(parsed_stats["Led"][0])
                return True if led == 1 else False
            except (IndexError, KeyError, ValueError, TypeError):
                pass

        try:
            data = await self.rpc.ascset(0, "led", "1-255")
        except APIError:
            return False
        try:
            if data["STATUS"][0]["Msg"] == "ASC 0 set info: LED[1]":
                return True
        except LookupError:
            pass
        return False



from pyasic.miners.backends import M5X
from pyasic.miners.device.models import M50SVH10


class BTMinerM50SVH10(M5X, M50SVH10):
    pass


from pyasic.miners.device.models import M50SVH20


class BTMinerM50SVH20(M5X, M50SVH20):
    pass


from pyasic.miners.device.models import M50SVH30


class BTMinerM50SVH30(M5X, M50SVH30):
    pass


from pyasic.miners.device.models import M50SVH40


class BTMinerM50SVH40(M5X, M50SVH40):
    pass


from pyasic.miners.device.models import M50SVH50


class BTMinerM50SVH50(M5X, M50SVH50):
    pass


from pyasic.miners.device.models import M50SVJ10


class BTMinerM50SVJ10(M5X, M50SVJ10):
    pass


from pyasic.miners.device.models import M50SVJ20


class BTMinerM50SVJ20(M5X, M50SVJ20):
    pass


from pyasic.miners.device.models import M50SVJ30


class BTMinerM50SVJ30(M5X, M50SVJ30):
    pass


from pyasic.miners.device.models import M50SVJ40


class BTMinerM50SVJ40(M5X, M50SVJ40):
    pass


from pyasic.miners.device.models import M50SVJ50


class BTMinerM50SVJ50(M5X, M50SVJ50):
    pass


from pyasic.miners.device.models import M50SVK10


class BTMinerM50SVK10(M5X, M50SVK10):
    pass


from pyasic.miners.device.models import M50SVK20


class BTMinerM50SVK20(M5X, M50SVK20):
    pass


from pyasic.miners.device.models import M50SVK30


class BTMinerM50SVK30(M5X, M50SVK30):
    pass


from pyasic.miners.device.models import M50SVK50


class BTMinerM50SVK50(M5X, M50SVK50):
    pass


from pyasic.miners.device.models import M50SVK60


class BTMinerM50SVK60(M5X, M50SVK60):
    pass


from pyasic.miners.device.models import M50SVK70


class BTMinerM50SVK70(M5X, M50SVK70):
    pass


from pyasic.miners.device.models import M50SVK80


class BTMinerM50SVK80(M5X, M50SVK80):
    pass


from pyasic.miners.device.models import M50SVL20


class BTMinerM50SVL20(M5X, M50SVL20):
    pass


from pyasic.miners.device.models import M50SVL30


class BTMinerM50SVL30(M5X, M50SVL30):
    pass



from .M60 import (
    M60VK6A,
    M60VK10,
    M60VK20,
    M60VK30,
    M60VK40,
    M60VL10,
    M60VL20,
    M60VL30,
    M60VL40,
    M60VL50,
)
from .M60S import (
    M60SVK10,
    M60SVK20,
    M60SVK30,
    M60SVK40,
    M60SVL10,
    M60SVL20,
    M60SVL30,
    M60SVL40,
    M60SVL50,
    M60SVL60,
    M60SVL70,
)
from .M60S_Plus import (
    M60SPlusVK30,
    M60SPlusVK40,
    M60SPlusVK50,
    M60SPlusVK60,
    M60SPlusVK70,
    M60SPlusVL10,
    M60SPlusVL30,
    M60SPlusVL40,
    M60SPlusVL50,
    M60SPlusVL60,
)
from .M60S_Plus_Plus import M60SPlusPlusVL30, M60SPlusPlusVL40
from .M61 import (
    M61VK10,
    M61VK20,
    M61VK30,
    M61VK40,
    M61VL10,
    M61VL30,
    M61VL40,
    M61VL50,
    M61VL60,
)
from .M61S import M61SVL10, M61SVL20, M61SVL30
from .M61S_Plus import M61SPlusVL30
from .M62S_Plus import M62SPlusVK30
from .M63 import M63VK10, M63VK20, M63VK30, M63VL10, M63VL30
from .M63S import M63SVK10, M63SVK20, M63SVK30, M63SVK60, M63SVL10, M63SVL50, M63SVL60
from .M63S_Plus import (
    M63SPlusVK30,
    M63SPlusVL10,
    M63SPlusVL20,
    M63SPlusVL30,
    M63SPlusVL50,
)
from .M63S_Plus_Plus import M63SPlusPlusVL20
from .M64 import M64VL30, M64VL40
from .M64S import M64SVL30
from .M65S import M65SVK20, M65SVL60
from .M65S_Plus import M65SPlusVK30
from .M66 import M66VK20, M66VK30, M66VL20, M66VL30
from .M66S import (
    M66SVK20,
    M66SVK30,
    M66SVK40,
    M66SVK50,
    M66SVK60,
    M66SVL10,
    M66SVL20,
    M66SVL30,
    M66SVL40,
    M66SVL50,
)
from .M66S_Plus import (
    M66SPlusVK30,
    M66SPlusVL10,
    M66SPlusVL20,
    M66SPlusVL30,
    M66SPlusVL40,
    M66SPlusVL60,
)
from .M66S_Plus_Plus import M66SPlusPlusVL20
from .M67S import M67SVK30



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M53SPlusPlusVK10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53SPlusPlusVK10

    expected_chips = 198
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M53SPlusPlusVK20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53SPlusPlusVK20

    expected_chips = 192
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M53SPlusPlusVK30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53SPlusPlusVK30

    expected_chips = 240
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M53SPlusPlusVK50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53SPlusPlusVK50

    expected_chips = 186
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M53SPlusPlusVL10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53SPlusPlusVL10

    expected_chips = 128
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M53SPlusPlusVL30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53SPlusPlusVL30

    expected_chips = 174
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256



from __future__ import annotations

from .hashrate.base import AlgoHashRateType, GenericHashrate
from .hashrate.unit.base import AlgoHashRateUnitType, GenericUnit


class MinerAlgoMeta(type):
    name: str

    def __str__(cls):
        return cls.name


class MinerAlgoType(metaclass=MinerAlgoMeta):
    hashrate: type[AlgoHashRateType]
    unit: type[AlgoHashRateUnitType]


class GenericAlgo(MinerAlgoType):
    hashrate: type[GenericHashrate] = GenericHashrate
    unit: type[GenericUnit] = GenericUnit

    name = "Generic (Unknown)"



from pyasic.miners.backends import M3X
from pyasic.miners.device.models import M30SPlusPlusV10


class BTMinerM30SPlusPlusV10(M3X, M30SPlusPlusV10):
    pass


from pyasic.miners.device.models import M30SPlusPlusV20


class BTMinerM30SPlusPlusV20(M3X, M30SPlusPlusV20):
    pass


from pyasic.miners.device.models import M30SPlusPlusVE30


class BTMinerM30SPlusPlusVE30(M3X, M30SPlusPlusVE30):
    pass


from pyasic.miners.device.models import M30SPlusPlusVE40


class BTMinerM30SPlusPlusVE40(M3X, M30SPlusPlusVE40):
    pass


from pyasic.miners.device.models import M30SPlusPlusVE50


class BTMinerM30SPlusPlusVE50(M3X, M30SPlusPlusVE50):
    pass


from pyasic.miners.device.models import M30SPlusPlusVF40


class BTMinerM30SPlusPlusVF40(M3X, M30SPlusPlusVF40):
    pass


from pyasic.miners.device.models import M30SPlusPlusVG30


class BTMinerM30SPlusPlusVG30(M3X, M30SPlusPlusVG30):
    pass


from pyasic.miners.device.models import M30SPlusPlusVG40


class BTMinerM30SPlusPlusVG40(M3X, M30SPlusPlusVG40):
    pass


from pyasic.miners.device.models import M30SPlusPlusVG50


class BTMinerM30SPlusPlusVG50(M3X, M30SPlusPlusVG50):
    pass


from pyasic.miners.device.models import M30SPlusPlusVH10


class BTMinerM30SPlusPlusVH10(M3X, M30SPlusPlusVH10):
    pass


from pyasic.miners.device.models import M30SPlusPlusVH100


class BTMinerM30SPlusPlusVH100(M3X, M30SPlusPlusVH100):
    pass


from pyasic.miners.device.models import M30SPlusPlusVH110


class BTMinerM30SPlusPlusVH110(M3X, M30SPlusPlusVH110):
    pass


from pyasic.miners.device.models import M30SPlusPlusVH20


class BTMinerM30SPlusPlusVH20(M3X, M30SPlusPlusVH20):
    pass


from pyasic.miners.device.models import M30SPlusPlusVH30


class BTMinerM30SPlusPlusVH30(M3X, M30SPlusPlusVH30):
    pass


from pyasic.miners.device.models import M30SPlusPlusVH40


class BTMinerM30SPlusPlusVH40(M3X, M30SPlusPlusVH40):
    pass


from pyasic.miners.device.models import M30SPlusPlusVH50


class BTMinerM30SPlusPlusVH50(M3X, M30SPlusPlusVH50):
    pass


from pyasic.miners.device.models import M30SPlusPlusVH60


class BTMinerM30SPlusPlusVH60(M3X, M30SPlusPlusVH60):
    pass


from pyasic.miners.device.models import M30SPlusPlusVH70


class BTMinerM30SPlusPlusVH70(M3X, M30SPlusPlusVH70):
    pass


from pyasic.miners.device.models import M30SPlusPlusVH80


class BTMinerM30SPlusPlusVH80(M3X, M30SPlusPlusVH80):
    pass


from pyasic.miners.device.models import M30SPlusPlusVH90


class BTMinerM30SPlusPlusVH90(M3X, M30SPlusPlusVH90):
    pass


from pyasic.miners.device.models import M30SPlusPlusVI30


class BTMinerM30SPlusPlusVI30(M3X, M30SPlusPlusVI30):
    pass


from pyasic.miners.device.models import M30SPlusPlusVJ20


class BTMinerM30SPlusPlusVJ20(M3X, M30SPlusPlusVJ20):
    pass


from pyasic.miners.device.models import M30SPlusPlusVJ30


class BTMinerM30SPlusPlusVJ30(M3X, M30SPlusPlusVJ30):
    pass


from pyasic.miners.device.models import M30SPlusPlusVJ50


class BTMinerM30SPlusPlusVJ50(M3X, M30SPlusPlusVJ50):
    pass


from pyasic.miners.device.models import M30SPlusPlusVJ60


class BTMinerM30SPlusPlusVJ60(M3X, M30SPlusPlusVJ60):
    pass


from pyasic.miners.device.models import M30SPlusPlusVJ70


class BTMinerM30SPlusPlusVJ70(M3X, M30SPlusPlusVJ70):
    pass


from pyasic.miners.device.models import M30SPlusPlusVK30


class BTMinerM30SPlusPlusVK30(M3X, M30SPlusPlusVK30):
    pass


from pyasic.miners.device.models import M30SPlusPlusVK40


class BTMinerM30SPlusPlusVK40(M3X, M30SPlusPlusVK40):
    pass



from pyasic.miners.backends import M5X
from pyasic.miners.device.models import M54SPlusPlusVK30


class BTMinerM54SPlusPlusVK30(M5X, M54SPlusPlusVK30):
    pass


from pyasic.miners.device.models import M54SPlusPlusVL30


class BTMinerM54SPlusPlusVL30(M5X, M54SPlusPlusVL30):
    pass


from pyasic.miners.device.models import M54SPlusPlusVL40


class BTMinerM54SPlusPlusVL40(M5X, M54SPlusPlusVL40):
    pass



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M66VK20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M66VK20

    expected_chips = 184
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M66VK30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M66VK30

    expected_chips = 192
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M66VL20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M66VL20

    expected_chips = 160
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M66VL30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M66VL30

    expected_chips = 168
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256



from pydantic import BaseModel


class BaseMinerError(BaseModel):
    @classmethod
    def fields(cls):
        return list(cls.model_fields.keys())

    def asdict(self) -> dict:
        return self.model_dump()

    def as_dict(self) -> dict:
        """Get this dataclass as a dictionary.

        Returns:
            A dictionary version of this class.
        """
        return self.asdict()

    def as_influxdb(self, root_key: str, level_delimiter: str = ".") -> str:
        field_data = []

        if self.error_code is not None:
            field_data.append(
                f"{root_key}{level_delimiter}error_code={self.error_code}"
            )
        if self.error_message is not None:
            field_data.append(
                f'{root_key}{level_delimiter}error_message="{self.error_message}"'
            )

        return ",".join(field_data)



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------
from __future__ import annotations

from typing import Any

from pydantic import BaseModel

from pyasic.device.algorithm.hashrate import AlgoHashRateType


class HashBoard(BaseModel):
    """A Dataclass to standardize hashboard data.

    Attributes:
        slot: The slot of the board as an int.
        hashrate: The hashrate of the board in TH/s as a float.
        inlet_temp: Inlet temperature for hydro asics as an int
        outlet_temp: Outlet temperature for hydro asics as an int
        temp: The temperature of the PCB as an int.
        chip_temp: The temperature of the chips as an int.
        chips: The chip count of the board as an int.
        expected_chips: The expected chip count of the board as an int.
        serial_number: The serial number of the board.
        missing: Whether the board is returned from the miners data as a bool.
        tuned: Whether the board is tuned as a bool.
        active: Whether the board is currently tuning as a bool.
        voltage: Current input voltage of the board as a float.
    """

    slot: int = 0
    hashrate: AlgoHashRateType | None = None
    inlet_temp: float | None = None
    outlet_temp: float | None = None
    temp: float | None = None
    chip_temp: float | None = None
    chips: int | None = None
    expected_chips: int | None = None
    serial_number: str | None = None
    missing: bool = True
    tuned: bool | None = None
    active: bool | None = None
    voltage: float | None = None

    @classmethod
    def fields(cls) -> set:
        all_fields = set(cls.model_fields.keys())
        all_fields.update(set(cls.model_computed_fields.keys()))
        return all_fields

    def get(self, __key: str, default: Any = None):
        try:
            val = self.__getitem__(__key)
            if val is None:
                return default
            return val
        except KeyError:
            return default

    def __getitem__(self, item: str):
        try:
            return getattr(self, item)
        except AttributeError:
            raise KeyError(f"{item}")

    def as_influxdb(self, key_root: str, level_delimiter: str = ".") -> str:

        def serialize_int(key: str, value: int) -> str:
            return f"{key}={value}"

        def serialize_float(key: str, value: float) -> str:
            return f"{key}={value}"

        def serialize_str(key: str, value: str) -> str:
            return f'{key}="{value}"'

        def serialize_algo_hash_rate(key: str, value: AlgoHashRateType) -> str:
            return f"{key}={round(float(value), 2)}"

        def serialize_bool(key: str, value: bool):
            return f"{key}={str(value).lower()}"

        serialization_map_instance = {
            AlgoHashRateType: serialize_algo_hash_rate,
        }
        serialization_map = {
            int: serialize_int,
            float: serialize_float,
            str: serialize_str,
            bool: serialize_bool,
        }

        include = [
            "hashrate",
            "temp",
            "chip_temp",
            "chips",
            "expected_chips",
            "tuned",
            "active",
            "voltage",
        ]

        field_data = []
        for field in include:
            field_val = getattr(self, field)
            serialization_func = serialization_map.get(
                type(field_val), lambda _k, _v: None
            )
            serialized = serialization_func(
                f"{key_root}{level_delimiter}{field}", field_val
            )
            if serialized is not None:
                field_data.append(serialized)
                continue
            for datatype in serialization_map_instance:
                if serialized is None:
                    if isinstance(field_val, datatype):
                        serialized = serialization_map_instance[datatype](
                            f"{key_root}{level_delimiter}{field}", field_val
                        )
            if serialized is not None:
                field_data.append(serialized)
        return ",".join(field_data)



from pyasic.miners.backends import M5X
from pyasic.miners.device.models import M53SPlusVJ30


class BTMinerM53SPlusVJ30(M5X, M53SPlusVJ30):
    pass


from pyasic.miners.device.models import M53SPlusVJ40


class BTMinerM53SPlusVJ40(M5X, M53SPlusVJ40):
    pass


from pyasic.miners.device.models import M53SPlusVJ50


class BTMinerM53SPlusVJ50(M5X, M53SPlusVJ50):
    pass


from pyasic.miners.device.models import M53SPlusVK30


class BTMinerM53SPlusVK30(M5X, M53SPlusVK30):
    pass



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M53SPlusVJ30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53SPlusVJ30

    expected_chips = 240
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M53SPlusVJ40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53SPlusVJ40

    expected_chips = 248
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M53SPlusVJ50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53SPlusVJ50

    expected_chips = 264
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M53SPlusVK30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53SPlusVK30

    expected_chips = 168
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M65SVK20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M65SVK20

    expected_chips = 350
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M65SVL60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M65SVL60

    expected_chips = 288
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M31SEV10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SEV10

    expected_chips = 82
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SEV20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SEV20

    expected_chips = 78
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M31SEV30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SEV30

    expected_chips = 78
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256



from pyasic.miners.backends import M3X
from pyasic.miners.device.models import M39V10


class BTMinerM39V10(M3X, M39V10):
    pass


from pyasic.miners.device.models import M39V20


class BTMinerM39V20(M3X, M39V20):
    pass


from pyasic.miners.device.models import M39V30


class BTMinerM39V30(M3X, M39V30):
    pass



from pyasic.miners.backends import M6X
from pyasic.miners.device.models import M60SPlusPlusVL30


class BTMinerM60SPlusPlusVL30(M6X, M60SPlusPlusVL30):
    pass


from pyasic.miners.device.models import M60SPlusPlusVL40


class BTMinerM60SPlusPlusVL40(M6X, M60SPlusPlusVL40):
    pass



from pyasic.miners.backends import M2X
from pyasic.miners.device.models import M21SV20


class BTMinerM21SV20(M2X, M21SV20):
    pass


from pyasic.miners.device.models import M21SV60


class BTMinerM21SV60(M2X, M21SV60):
    pass


from pyasic.miners.device.models import M21SV70


class BTMinerM21SV70(M2X, M21SV70):
    pass



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M33SPlusPlusVG40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M33SPlusPlusVG40

    expected_chips = 174
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M33SPlusPlusVH20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M33SPlusPlusVH20

    expected_chips = 112
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M33SPlusPlusVH30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M33SPlusPlusVH30

    expected_chips = None
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------

from pathlib import Path
from typing import List, Optional

from pyasic.config import MinerConfig
from pyasic.data import Fan, HashBoard
from pyasic.data.error_codes import MinerErrorData, X19Error
from pyasic.data.pools import PoolMetrics, PoolUrl
from pyasic.device.algorithm import AlgoHashRate, ScryptAlgo
from pyasic.errors import APIError
from pyasic.logger import logger
from pyasic.miners.data import DataFunction, DataLocations, DataOptions, WebAPICommand
from pyasic.miners.device.firmware import ePICFirmware
from pyasic.web.epic import ePICWebAPI

EPIC_DATA_LOC = DataLocations(
    **{
        str(DataOptions.MAC): DataFunction(
            "_get_mac",
            [WebAPICommand("web_network", "network")],
        ),
        str(DataOptions.FW_VERSION): DataFunction(
            "_get_fw_ver",
            [WebAPICommand("web_summary", "summary")],
        ),
        str(DataOptions.HOSTNAME): DataFunction(
            "_get_hostname",
            [WebAPICommand("web_summary", "summary")],
        ),
        str(DataOptions.HASHRATE): DataFunction(
            "_get_hashrate",
            [WebAPICommand("web_summary", "summary")],
        ),
        str(DataOptions.EXPECTED_HASHRATE): DataFunction(
            "_get_expected_hashrate",
            [WebAPICommand("web_summary", "summary")],
        ),
        str(DataOptions.HASHBOARDS): DataFunction(
            "_get_hashboards",
            [
                WebAPICommand("web_summary", "summary"),
                WebAPICommand("web_capabilities", "capabilities"),
            ],
        ),
        str(DataOptions.WATTAGE): DataFunction(
            "_get_wattage",
            [WebAPICommand("web_summary", "summary")],
        ),
        str(DataOptions.FANS): DataFunction(
            "_get_fans",
            [WebAPICommand("web_summary", "summary")],
        ),
        str(DataOptions.ERRORS): DataFunction(
            "_get_errors",
            [WebAPICommand("web_summary", "summary")],
        ),
        str(DataOptions.FAULT_LIGHT): DataFunction(
            "_get_fault_light",
            [WebAPICommand("web_summary", "summary")],
        ),
        str(DataOptions.UPTIME): DataFunction(
            "_get_uptime",
            [WebAPICommand("web_summary", "summary")],
        ),
        str(DataOptions.IS_MINING): DataFunction(
            "_is_mining",
            [WebAPICommand("web_summary", "summary")],
        ),
        str(DataOptions.POOLS): DataFunction(
            "_get_pools",
            [WebAPICommand("web_summary", "summary")],
        ),
    }
)


class ePIC(ePICFirmware):
    """Handler for miners with the ePIC board"""

    _web_cls = ePICWebAPI
    web: ePICWebAPI

    data_locations = EPIC_DATA_LOC

    supports_shutdown = True

    async def get_config(self) -> MinerConfig:
        summary = None
        try:
            summary = await self.web.summary()
        except APIError as e:
            logger.warning(e)
        except LookupError:
            pass

        if summary is not None:
            cfg = MinerConfig.from_epic(summary)
        else:
            cfg = MinerConfig()

        self.config = cfg
        return self.config

    async def send_config(self, config: MinerConfig, user_suffix: str = None) -> None:
        self.config = config
        conf = self.config.as_epic(user_suffix=user_suffix)

        try:
            # Temps
            if not conf.get("temps", {}) == {}:
                await self.web.set_shutdown_temp(conf["temps"]["shutdown"])
                await self.web.set_critical_temp(conf["temps"]["critical"])
            # Fans
            # set with sub-keys instead of conf["fans"] because sometimes both can be set
            if not conf["fans"].get("Manual", {}) == {}:
                await self.web.set_fan({"Manual": conf["fans"]["Manual"]})
            elif not conf["fans"].get("Auto", {}) == {}:
                await self.web.set_fan({"Auto": conf["fans"]["Auto"]})

            # Mining Mode -- Need to handle that you may not be able to change while miner is tuning
            if conf["ptune"].get("enabled", True):
                await self.web.set_ptune_enable(True)
                await self.web.set_ptune_algo(conf["ptune"])

            ## Pools
            if self.algo == ScryptAlgo:
                conf["pools"]["coin"] = "Ltc"
            await self.web.set_pools(conf["pools"])
        except APIError:
            pass

    async def restart_backend(self) -> bool:
        data = await self.web.restart_epic()
        if data:
            try:
                return data["success"]
            except KeyError:
                pass
        return False

    async def stop_mining(self) -> bool:
        data = await self.web.stop_mining()
        if data:
            try:
                return data["success"]
            except KeyError:
                pass
        return False

    async def resume_mining(self) -> bool:
        data = await self.web.resume_mining()
        if data:
            try:
                return data["success"]
            except KeyError:
                pass
        return False

    async def reboot(self) -> bool:
        data = await self.web.reboot()
        if data:
            try:
                return data["success"]
            except KeyError:
                pass
        return False

    async def _get_mac(self, web_network: dict = None) -> Optional[str]:
        if web_network is None:
            try:
                web_network = await self.web.network()
            except APIError:
                pass

        if web_network is not None:
            try:
                for network in web_network:
                    mac = web_network[network]["mac_address"]
                    return mac
            except KeyError:
                pass

    async def _get_hostname(self, web_summary: dict = None) -> Optional[str]:
        if web_summary is None:
            try:
                web_summary = await self.web.summary()
            except APIError:
                pass

        if web_summary is not None:
            try:
                hostname = web_summary["Hostname"]
                return hostname
            except KeyError:
                pass

    async def _get_wattage(self, web_summary: dict = None) -> Optional[int]:
        if web_summary is None:
            try:
                web_summary = await self.web.summary()
            except APIError:
                pass

        if web_summary is not None:
            try:
                wattage = web_summary["Power Supply Stats"]["Input Power"]
                wattage = round(wattage)
                return wattage
            except KeyError:
                pass

    async def _get_hashrate(self, web_summary: dict = None) -> Optional[AlgoHashRate]:
        if web_summary is None:
            try:
                web_summary = await self.web.summary()
            except APIError:
                pass

        if web_summary is not None:
            try:
                hashrate = 0
                if web_summary["HBs"] is not None:
                    for hb in web_summary["HBs"]:
                        hashrate += hb["Hashrate"][0]
                    return self.algo.hashrate(
                        rate=float(hashrate), unit=self.algo.unit.MH
                    ).into(self.algo.unit.TH)
            except (LookupError, ValueError, TypeError):
                pass

    async def _get_expected_hashrate(
        self, web_summary: dict = None
    ) -> Optional[AlgoHashRate]:
        if web_summary is None:
            try:
                web_summary = await self.web.summary()
            except APIError:
                pass

        if web_summary is not None:
            try:
                hashrate = 0
                if web_summary.get("HBs") is not None:
                    for hb in web_summary["HBs"]:
                        if hb["Hashrate"][1] == 0:
                            ideal = 1.0
                        else:
                            ideal = hb["Hashrate"][1] / 100

                        hashrate += hb["Hashrate"][0] / ideal
                    return self.algo.hashrate(
                        rate=float(hashrate), unit=self.algo.unit.MH
                    ).into(self.algo.unit.default)
            except (LookupError, ValueError, TypeError):
                pass

    async def _get_fw_ver(self, web_summary: dict = None) -> Optional[str]:
        if web_summary is None:
            try:
                web_summary = await self.web.summary()
            except APIError:
                pass

        if web_summary is not None:
            try:
                fw_ver = web_summary["Software"]
                fw_ver = fw_ver.split(" ")[1].replace("v", "")
                return fw_ver
            except KeyError:
                pass

    async def _get_fans(self, web_summary: dict = None) -> List[Fan]:
        if self.expected_fans is None:
            return []

        if web_summary is None:
            try:
                web_summary = await self.web.summary()
            except APIError:
                pass

        fans = []

        if web_summary is not None:
            for fan in web_summary["Fans Rpm"]:
                try:
                    fans.append(Fan(speed=web_summary["Fans Rpm"][fan]))
                except (LookupError, ValueError, TypeError):
                    fans.append(Fan())
        return fans

    async def _get_hashboards(
        self, web_summary: dict = None, web_capabilities: dict = None
    ) -> List[HashBoard]:
        if self.expected_hashboards is None:
            return []

        if web_summary is None:
            try:
                web_summary = await self.web.summary()
            except APIError:
                pass

        if web_capabilities is None:
            try:
                web_capabilities = await self.web.capabilities()
            except APIError:
                pass

        tuned = True
        active = False
        if web_summary is not None:
            tuner_running = web_summary["PerpetualTune"]["Running"]
            if tuner_running:
                active = True
                algo_info = web_summary["PerpetualTune"]["Algorithm"]
                if algo_info.get("VoltageOptimizer") is not None:
                    tuned = algo_info["VoltageOptimizer"].get("Optimized")
                elif algo_info.get("BoardTune") is not None:
                    tuned = algo_info["BoardTune"].get("Optimized")
                else:
                    tuned = algo_info["ChipTune"].get("Optimized")

            # To be extra detailed, also ensure the miner is in "Mining" state
            tuned = tuned and web_summary["Status"]["Operating State"] == "Mining"
            active = active and web_summary["Status"]["Operating State"] == "Mining"

        hb_list = [
            HashBoard(slot=i, expected_chips=self.expected_chips)
            for i in range(self.expected_hashboards)
        ]
        if web_summary is not None and web_capabilities is not None:
            if web_summary.get("HBs") is not None:
                for hb in web_summary["HBs"]:
                    if web_capabilities.get("Performance Estimator") is not None:
                        num_of_chips = web_capabilities["Performance Estimator"][
                            "Chip Count"
                        ]
                    else:
                        num_of_chips = self.expected_chips
                    if web_capabilities.get("Board Serial Numbers") is not None:
                        if hb["Index"] < len(web_capabilities["Board Serial Numbers"]):
                            hb_list[hb["Index"]].serial_number = web_capabilities[
                                "Board Serial Numbers"
                            ][hb["Index"]]
                    hashrate = hb["Hashrate"][0]
                    # Update the Hashboard object
                    hb_list[hb["Index"]].missing = False
                    hb_list[hb["Index"]].hashrate = self.algo.hashrate(
                        rate=float(hashrate), unit=self.algo.unit.MH
                    ).into(self.algo.unit.default)
                    hb_list[hb["Index"]].chips = num_of_chips
                    hb_list[hb["Index"]].temp = int(hb["Temperature"])
                    hb_list[hb["Index"]].tuned = tuned
                    hb_list[hb["Index"]].active = active
                    hb_list[hb["Index"]].voltage = hb["Input Voltage"]
            return hb_list

    async def _is_mining(self, web_summary, *args, **kwargs) -> Optional[bool]:
        if web_summary is None:
            try:
                web_summary = await self.web.summary()
            except APIError:
                pass
        if web_summary is not None:
            try:
                op_state = web_summary["Status"]["Operating State"]
                return not op_state == "Idling"
            except KeyError:
                pass

    async def _get_uptime(self, web_summary: dict = None) -> Optional[int]:
        if web_summary is None:
            try:
                web_summary = await self.web.summary()
            except APIError:
                pass

        if web_summary is not None:
            try:
                uptime = web_summary["Session"]["Uptime"]
                return uptime
            except KeyError:
                pass
        return None

    async def _get_fault_light(self, web_summary: dict = None) -> Optional[bool]:
        if web_summary is None:
            try:
                web_summary = await self.web.summary()
            except APIError:
                pass

        if web_summary is not None:
            try:
                light = web_summary["Misc"]["Locate Miner State"]
                return light
            except KeyError:
                pass
        return False

    async def _get_errors(self, web_summary: dict = None) -> List[MinerErrorData]:
        if not web_summary:
            try:
                web_summary = await self.web.summary()
            except APIError:
                pass

        errors = []
        if web_summary is not None:
            try:
                error = web_summary["Status"]["Last Error"]
                if error is not None:
                    errors.append(X19Error(error_message=str(error)))
                return errors
            except KeyError:
                pass
        return errors

    async def _get_pools(self, web_summary: dict = None) -> List[PoolMetrics]:
        if web_summary is None:
            try:
                web_summary = await self.web.summary()
            except APIError:
                pass

        pool_data = []
        try:
            if web_summary is not None:
                if (
                    web_summary.get("Session") is not None
                    and web_summary.get("Stratum") is not None
                ):
                    url = web_summary["Stratum"].get("Current Pool")
                    # TODO: when scheme gets put in, update this
                    if url is not None:
                        url = PoolUrl.from_str(f"stratum+tcp://{url}")
                    pool_data.append(
                        PoolMetrics(
                            accepted=web_summary["Session"].get("Accepted"),
                            rejected=web_summary["Session"].get("Rejected"),
                            get_failures=0,
                            remote_failures=0,
                            active=web_summary["Stratum"].get("IsPoolConnected"),
                            alive=web_summary["Stratum"].get("IsPoolConnected"),
                            url=url,
                            user=web_summary["Stratum"].get("Current User"),
                            index=web_summary["Stratum"].get("Config Id"),
                        )
                    )
                return pool_data
        except LookupError:
            pass

    async def upgrade_firmware(
        self, file: Path | str, keep_settings: bool = True
    ) -> bool:
        """
        Upgrade the firmware of the ePIC miner device.

        Args:
            file (Path | str): The local file path of the firmware to be uploaded.
            keep_settings (bool): Whether to keep the current settings after the update.

        Returns:
            bool: Whether the firmware update succeeded.
        """
        return await self.web.system_update(file=file, keep_settings=keep_settings)



from __future__ import annotations

from .base import AlgoHashRateUnitType


class EtHashUnit(AlgoHashRateUnitType):
    H = 1
    KH = int(H) * 1000
    MH = int(KH) * 1000
    GH = int(MH) * 1000
    TH = int(GH) * 1000
    PH = int(TH) * 1000
    EH = int(PH) * 1000
    ZH = int(EH) * 1000

    default = MH



from typing import List, Optional

from pyasic import MinerConfig
from pyasic.data import Fan, HashBoard
from pyasic.data.pools import PoolMetrics, PoolUrl
from pyasic.device.algorithm import AlgoHashRate, MinerAlgo
from pyasic.errors import APIError
from pyasic.miners.data import DataFunction, DataLocations, DataOptions, WebAPICommand
from pyasic.miners.device.firmware import StockFirmware
from pyasic.web.iceriver import IceRiverWebAPI

ICERIVER_DATA_LOC = DataLocations(
    **{
        str(DataOptions.MAC): DataFunction(
            "_get_mac",
            [WebAPICommand("web_userpanel", "userpanel")],
        ),
        str(DataOptions.FANS): DataFunction(
            "_get_fans",
            [WebAPICommand("web_userpanel", "userpanel")],
        ),
        str(DataOptions.HOSTNAME): DataFunction(
            "_get_hostname",
            [WebAPICommand("web_userpanel", "userpanel")],
        ),
        str(DataOptions.HASHRATE): DataFunction(
            "_get_hashrate",
            [WebAPICommand("web_userpanel", "userpanel")],
        ),
        str(DataOptions.IS_MINING): DataFunction(
            "_is_mining",
            [WebAPICommand("web_userpanel", "userpanel")],
        ),
        str(DataOptions.FAULT_LIGHT): DataFunction(
            "_get_fault_light",
            [WebAPICommand("web_userpanel", "userpanel")],
        ),
        str(DataOptions.HASHBOARDS): DataFunction(
            "_get_hashboards",
            [WebAPICommand("web_userpanel", "userpanel")],
        ),
        str(DataOptions.UPTIME): DataFunction(
            "_get_uptime",
            [WebAPICommand("web_userpanel", "userpanel")],
        ),
        str(DataOptions.POOLS): DataFunction(
            "_get_pools",
            [WebAPICommand("web_userpanel", "userpanel")],
        ),
    }
)


class IceRiver(StockFirmware):
    """Handler for IceRiver miners"""

    _web_cls = IceRiverWebAPI
    web: IceRiverWebAPI

    data_locations = ICERIVER_DATA_LOC

    async def fault_light_off(self) -> bool:
        try:
            await self.web.locate(False)
        except APIError:
            return False
        return True

    async def fault_light_on(self) -> bool:
        try:
            await self.web.locate(True)
        except APIError:
            return False
        return True

    async def get_config(self) -> MinerConfig:
        web_userpanel = await self.web.userpanel()

        return MinerConfig.from_iceriver(web_userpanel)

    async def _get_fans(self, web_userpanel: dict = None) -> List[Fan]:
        if self.expected_fans is None:
            return []

        if web_userpanel is None:
            try:
                web_userpanel = await self.web.userpanel()
            except APIError:
                pass

        if web_userpanel is not None:
            try:
                return [
                    Fan(speed=spd) for spd in web_userpanel["userpanel"]["data"]["fans"]
                ]
            except (LookupError, ValueError, TypeError):
                pass

    async def _get_mac(self, web_userpanel: dict = None) -> Optional[str]:
        if web_userpanel is None:
            try:
                web_userpanel = await self.web.userpanel()
            except APIError:
                pass

        if web_userpanel is not None:
            try:
                return (
                    web_userpanel["userpanel"]["data"]["mac"].upper().replace("-", ":")
                )
            except (LookupError, ValueError, TypeError):
                pass

    async def _get_hostname(self, web_userpanel: dict = None) -> Optional[str]:
        if web_userpanel is None:
            try:
                web_userpanel = await self.web.userpanel()
            except APIError:
                pass

        if web_userpanel is not None:
            try:
                return web_userpanel["userpanel"]["data"]["host"]
            except (LookupError, ValueError, TypeError):
                pass

    async def _get_hashrate(self, web_userpanel: dict = None) -> Optional[AlgoHashRate]:
        if web_userpanel is None:
            try:
                web_userpanel = await self.web.userpanel()
            except APIError:
                pass

        if web_userpanel is not None:
            try:
                base_unit = web_userpanel["userpanel"]["data"]["unit"]
                return self.algo.hashrate(
                    rate=float(
                        web_userpanel["userpanel"]["data"]["rtpow"].replace(
                            base_unit, ""
                        )
                    ),
                    unit=MinerAlgo.SHA256.unit.from_str(base_unit + "H"),
                ).into(MinerAlgo.SHA256.unit.default)
            except (LookupError, ValueError, TypeError):
                pass

    async def _get_fault_light(self, web_userpanel: dict = None) -> bool:
        if web_userpanel is None:
            try:
                web_userpanel = await self.web.userpanel()
            except APIError:
                pass

        if web_userpanel is not None:
            try:
                return web_userpanel["userpanel"]["data"]["locate"]
            except (LookupError, ValueError, TypeError):
                pass
        return False

    async def _is_mining(self, web_userpanel: dict = None) -> Optional[bool]:
        if web_userpanel is None:
            try:
                web_userpanel = await self.web.userpanel()
            except APIError:
                pass

        if web_userpanel is not None:
            try:
                return web_userpanel["userpanel"]["data"]["powstate"]
            except (LookupError, ValueError, TypeError):
                pass

    async def _get_hashboards(self, web_userpanel: dict = None) -> List[HashBoard]:
        if self.expected_hashboards is None:
            return []

        if web_userpanel is None:
            try:
                web_userpanel = await self.web.userpanel()
            except APIError:
                pass

        hb_list = [
            HashBoard(slot=i, expected_chips=self.expected_chips)
            for i in range(self.expected_hashboards)
        ]

        if web_userpanel is not None:
            try:
                for board in web_userpanel["userpanel"]["data"]["boards"]:
                    idx = int(board["no"] - 1)
                    hb_list[idx].chip_temp = round(board["outtmp"])
                    hb_list[idx].temp = round(board["intmp"])
                    hb_list[idx].hashrate = self.algo.hashrate(
                        rate=float(board["rtpow"].replace("G", "")),
                        unit=self.algo.unit.GH,
                    ).into(self.algo.unit.default)
                    hb_list[idx].chips = int(board["chipnum"])
                    hb_list[idx].missing = False
            except LookupError:
                pass
        return hb_list

    async def _get_uptime(self, web_userpanel: dict = None) -> Optional[int]:
        if web_userpanel is None:
            try:
                web_userpanel = await self.web.userpanel()
            except APIError:
                pass

        if web_userpanel is not None:
            try:
                runtime = web_userpanel["userpanel"]["data"]["runtime"]
                days, hours, minutes, seconds = runtime.split(":")
                return (
                    (int(days) * 24 * 60 * 60)
                    + (int(hours) * 60 * 60)
                    + (int(minutes) * 60)
                    + int(seconds)
                )
            except (LookupError, ValueError, TypeError):
                pass

    async def _get_pools(self, web_userpanel: dict = None) -> List[PoolMetrics]:
        if web_userpanel is None:
            try:
                web_userpanel = await self.web.userpanel()
            except APIError:
                pass

        pools_data = []
        if web_userpanel is not None:
            try:
                pools = web_userpanel["userpanel"]["data"]["pools"]
                for pool_info in pools:
                    pool_num = pool_info.get("no")
                    if pool_num is not None:
                        pool_num = int(pool_num)
                    if pool_info["addr"] == "":
                        continue
                    url = pool_info.get("addr")
                    pool_url = PoolUrl.from_str(url) if url else None
                    pool_data = PoolMetrics(
                        accepted=pool_info.get("accepted"),
                        rejected=pool_info.get("rejected"),
                        active=pool_info.get("connect"),
                        alive=int(pool_info.get("state", 0)) == 1,
                        url=pool_url,
                        user=pool_info.get("user"),
                        index=pool_num,
                    )
                    pools_data.append(pool_data)
            except LookupError:
                pass
        return pools_data



from __future__ import annotations

import asyncio
import json
from typing import Any

import httpx

from pyasic import APIError, settings
from pyasic.web.base import BaseWebAPI


class ESPMinerWebAPI(BaseWebAPI):
    async def send_command(
        self,
        command: str | bytes,
        ignore_errors: bool = False,
        allow_warning: bool = True,
        privileged: bool = False,
        **parameters: Any,
    ) -> dict:
        url = f"http://{self.ip}:{self.port}/api/{command}"
        async with httpx.AsyncClient(transport=settings.transport()) as client:
            for _ in range(settings.get("get_data_retries", 1)):
                try:
                    if parameters.get("post", False):
                        parameters.pop("post")
                        data = await client.post(
                            url,
                            timeout=settings.get("api_function_timeout", 3),
                            json=parameters,
                        )
                    elif parameters.get("patch", False):
                        parameters.pop("patch")
                        data = await client.patch(
                            url,
                            timeout=settings.get("api_function_timeout", 3),
                            json=parameters,
                        )
                    else:
                        data = await client.get(
                            url,
                            timeout=settings.get("api_function_timeout", 5),
                        )
                except httpx.HTTPError:
                    pass
                else:
                    if data.status_code == 200:
                        try:
                            return data.json()
                        except json.decoder.JSONDecodeError:
                            pass

    async def multicommand(
        self, *commands: str, ignore_errors: bool = False, allow_warning: bool = True
    ) -> dict:
        """Execute multiple commands simultaneously on the BitAxe miner.

        Args:
            *commands (str): Commands to execute.
            ignore_errors (bool): Whether to ignore errors during command execution.
            allow_warning (bool): Whether to proceed despite warnings.

        Returns:
            dict: A dictionary containing responses for all commands executed.
        """
        tasks = {}
        # send all commands individually
        for cmd in commands:
            tasks[cmd] = asyncio.create_task(
                self.send_command(cmd, allow_warning=allow_warning)
            )

        await asyncio.gather(*[tasks[cmd] for cmd in tasks], return_exceptions=True)

        data = {"multicommand": True}
        for cmd in tasks:
            try:
                result = tasks[cmd].result()
                if result is None or result == {}:
                    result = {}
                data[cmd] = result
            except APIError:
                pass

        return data

    async def system_info(self):
        return await self.send_command("system/info")

    async def swarm_info(self):
        return await self.send_command("swarm/info")

    async def restart(self):
        return await self.send_command("system/restart", post=True)

    async def update_settings(self, **config):
        return await self.send_command("system", patch=True, **config)



from pyasic.miners.backends import M3X
from pyasic.miners.device.models import M31HV10


class BTMinerM31HV10(M3X, M31HV10):
    pass


from pyasic.miners.device.models import M31HV40


class BTMinerM31HV40(M3X, M31HV40):
    pass



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------
import logging
from typing import List, Optional

from pyasic.config import MinerConfig
from pyasic.config.mining import MiningModePreset
from pyasic.data import Fan, HashBoard
from pyasic.data.pools import PoolMetrics, PoolUrl
from pyasic.device.algorithm import AlgoHashRate
from pyasic.errors import APIError
from pyasic.miners.data import DataFunction, DataLocations, DataOptions, RPCAPICommand
from pyasic.miners.device.firmware import LuxOSFirmware
from pyasic.rpc.luxminer import LUXMinerRPCAPI

LUXMINER_DATA_LOC = DataLocations(
    **{
        str(DataOptions.MAC): DataFunction(
            "_get_mac",
            [RPCAPICommand("rpc_config", "config")],
        ),
        str(DataOptions.HASHRATE): DataFunction(
            "_get_hashrate",
            [RPCAPICommand("rpc_summary", "summary")],
        ),
        str(DataOptions.EXPECTED_HASHRATE): DataFunction(
            "_get_expected_hashrate",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.HASHBOARDS): DataFunction(
            "_get_hashboards",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.WATTAGE): DataFunction(
            "_get_wattage",
            [RPCAPICommand("rpc_power", "power")],
        ),
        str(DataOptions.WATTAGE_LIMIT): DataFunction(
            "_get_wattage_limit",
            [
                RPCAPICommand("rpc_config", "config"),
                RPCAPICommand("rpc_profiles", "profiles"),
            ],
        ),
        str(DataOptions.FANS): DataFunction(
            "_get_fans",
            [RPCAPICommand("rpc_fans", "fans")],
        ),
        str(DataOptions.UPTIME): DataFunction(
            "_get_uptime", [RPCAPICommand("rpc_stats", "stats")]
        ),
        str(DataOptions.POOLS): DataFunction(
            "_get_pools", [RPCAPICommand("rpc_pools", "pools")]
        ),
        str(DataOptions.FW_VERSION): DataFunction(
            "_get_fw_ver", [RPCAPICommand("rpc_version", "version")]
        ),
        str(DataOptions.API_VERSION): DataFunction(
            "_get_api_ver", [RPCAPICommand("rpc_version", "version")]
        ),
        str(DataOptions.FAULT_LIGHT): DataFunction(
            "_get_fault_light", [RPCAPICommand("rpc_config", "config")]
        ),
    }
)


class LUXMiner(LuxOSFirmware):
    """Handler for LuxOS miners"""

    _rpc_cls = LUXMinerRPCAPI
    rpc: LUXMinerRPCAPI

    supports_shutdown = True
    supports_presets = True
    supports_autotuning = True

    data_locations = LUXMINER_DATA_LOC

    async def fault_light_on(self) -> bool:
        try:
            await self.rpc.ledset("red", "blink")
            return True
        except (APIError, LookupError):
            pass
        return False

    async def fault_light_off(self) -> bool:
        try:
            await self.rpc.ledset("red", "off")
            return True
        except (APIError, LookupError):
            pass
        return False

    async def restart_backend(self) -> bool:
        return await self.restart_luxminer()

    async def restart_luxminer(self) -> bool:
        try:
            await self.rpc.resetminer()
            return True
        except (APIError, LookupError):
            pass
        return False

    async def stop_mining(self) -> bool:
        try:
            await self.rpc.sleep()
            return True
        except (APIError, LookupError):
            pass
        return False

    async def resume_mining(self) -> bool:
        try:
            await self.rpc.wakeup()
            return True
        except (APIError, LookupError):
            pass

    async def reboot(self) -> bool:
        try:
            await self.rpc.rebootdevice()
            return True
        except (APIError, LookupError):
            pass
        return False

    async def get_config(self) -> MinerConfig:
        data = await self.rpc.multicommand(
            "tempctrl", "fans", "pools", "groups", "config", "profiles"
        )
        return MinerConfig.from_luxos(
            rpc_tempctrl=data.get("tempctrl", [{}])[0],
            rpc_fans=data.get("fans", [{}])[0],
            rpc_pools=data.get("pools", [{}])[0],
            rpc_groups=data.get("groups", [{}])[0],
            rpc_config=data.get("config", [{}])[0],
            rpc_profiles=data.get("profiles", [{}])[0],
        )

    async def upgrade_firmware(self, *args, **kwargs) -> bool:
        """
        Upgrade the firmware on a LuxOS miner by calling the 'updaterun' API command.
        Returns:
            bool: True if the firmware upgrade was successfully initiated, False otherwise.
        """
        try:
            await self.rpc.updaterun()
            logging.info(f"{self.ip}: Firmware upgrade initiated successfully.")
            return True

        except APIError as e:
            logging.error(f"{self.ip}: Firmware upgrade failed: {e}")

        return False

    async def atm_enabled(self) -> Optional[bool]:
        try:
            result = await self.rpc.atm()
            return result["ATM"][0]["Enabled"]
        except (APIError, LookupError):
            pass

    async def set_power_limit(self, wattage: int) -> bool:
        config = await self.get_config()
        valid_presets = {
            preset.name: preset.power
            for preset in config.mining_mode.available_presets
            if preset.power <= wattage
        }

        # Set power to highest preset <= wattage
        # If ATM enabled, must disable it before setting power limit
        new_preset = max(valid_presets, key=valid_presets.get)

        re_enable_atm = False
        try:
            if await self.atm_enabled():
                re_enable_atm = True
                await self.rpc.atmset(enabled=False)
            result = await self.rpc.profileset(new_preset)
            if re_enable_atm:
                await self.rpc.atmset(enabled=True)
        except APIError:
            raise
        except Exception as e:
            logging.warning(f"{self} - Failed to set power limit: {e}")
            return False

        if result["PROFILE"][0]["Profile"] == new_preset:
            return True
        else:
            return False

    ##################################################
    ### DATA GATHERING FUNCTIONS (get_{some_data}) ###
    ##################################################

    async def _get_mac(self, rpc_config: dict = None) -> Optional[str]:
        if rpc_config is None:
            try:
                rpc_config = await self.rpc.config()
            except APIError:
                pass

        if rpc_config is not None:
            try:
                return rpc_config["CONFIG"][0]["MACAddr"].upper()
            except KeyError:
                pass

    async def _get_hashrate(self, rpc_summary: dict = None) -> Optional[AlgoHashRate]:
        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        if rpc_summary is not None:
            try:
                return self.algo.hashrate(
                    rate=float(rpc_summary["SUMMARY"][0]["GHS 5s"]),
                    unit=self.algo.unit.GH,
                ).into(self.algo.unit.default)
            except (LookupError, ValueError, TypeError):
                pass

    async def _get_hashboards(self, rpc_stats: dict = None) -> List[HashBoard]:
        if self.expected_hashboards is None:
            return []

        hashboards = [
            HashBoard(slot=idx, expected_chips=self.expected_chips)
            for idx in range(self.expected_hashboards)
        ]

        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass
        if rpc_stats is not None:
            try:
                # TODO: bugged on S9 because of index issues, fix later.
                board_stats = rpc_stats["STATS"][1]
                for idx in range(3):
                    board_n = idx + 1
                    hashboards[idx].hashrate = self.algo.hashrate(
                        rate=float(board_stats[f"chain_rate{board_n}"]),
                        unit=self.algo.unit.GH,
                    ).into(self.algo.unit.default)
                    hashboards[idx].chips = int(board_stats[f"chain_acn{board_n}"])
                    chip_temp_data = list(
                        filter(
                            lambda x: not x == 0,
                            map(int, board_stats[f"temp_chip{board_n}"].split("-")),
                        )
                    )
                    hashboards[idx].chip_temp = (
                        sum([chip_temp_data[0], chip_temp_data[3]]) / 2
                    )
                    board_temp_data = list(
                        filter(
                            lambda x: not x == 0,
                            map(int, board_stats[f"temp_pcb{board_n}"].split("-")),
                        )
                    )
                    hashboards[idx].temp = (
                        sum([board_temp_data[1], board_temp_data[2]]) / 2
                    )
                    hashboards[idx].missing = False
            except LookupError:
                pass
        return hashboards

    async def _get_wattage(self, rpc_power: dict = None) -> Optional[int]:
        if rpc_power is None:
            try:
                rpc_power = await self.rpc.power()
            except APIError:
                pass

        if rpc_power is not None:
            try:
                return rpc_power["POWER"][0]["Watts"]
            except (LookupError, ValueError, TypeError):
                pass

    async def _get_wattage_limit(
        self, rpc_config: dict = None, rpc_profiles: list[dict] = None
    ) -> Optional[int]:
        try:
            active_preset = MiningModePreset.get_active_preset_from_luxos(
                rpc_config, rpc_profiles
            )
            return active_preset.power
        except (LookupError, ValueError, TypeError):
            pass

    async def _get_fans(self, rpc_fans: dict = None) -> List[Fan]:
        if self.expected_fans is None:
            return []

        if rpc_fans is None:
            try:
                rpc_fans = await self.rpc.fans()
            except APIError:
                pass

        fans = []

        if rpc_fans is not None:
            for fan in range(self.expected_fans):
                try:
                    fans.append(Fan(speed=rpc_fans["FANS"][fan]["RPM"]))
                except (LookupError, ValueError, TypeError):
                    fans.append(Fan())
        return fans

    async def _get_expected_hashrate(
        self, rpc_stats: dict = None
    ) -> Optional[AlgoHashRate]:
        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        if rpc_stats is not None:
            try:
                expected_rate = rpc_stats["STATS"][1]["total_rateideal"]
                try:
                    rate_unit = rpc_stats["STATS"][1]["rate_unit"]
                except KeyError:
                    rate_unit = "GH"
                return self.algo.hashrate(
                    rate=float(expected_rate), unit=self.algo.unit.from_str(rate_unit)
                ).into(self.algo.unit.default)
            except LookupError:
                pass

    async def _get_uptime(self, rpc_stats: dict = None) -> Optional[int]:
        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        if rpc_stats is not None:
            try:
                return int(rpc_stats["STATS"][1]["Elapsed"])
            except LookupError:
                pass

    async def _get_fw_ver(self, rpc_version: dict = None) -> Optional[str]:
        if rpc_version is None:
            try:
                rpc_version = await self.rpc.version()
            except APIError:
                pass

        if rpc_version is not None:
            try:
                return rpc_version["VERSION"][0]["Miner"]
            except LookupError:
                pass

    async def _get_api_ver(self, rpc_version: dict = None) -> Optional[str]:
        if rpc_version is None:
            try:
                rpc_version = await self.rpc.version()
            except APIError:
                pass

        if rpc_version is not None:
            try:
                return rpc_version["VERSION"][0]["API"]
            except LookupError:
                pass

    async def _get_fault_light(self, rpc_config: dict = None) -> Optional[bool]:
        if rpc_config is None:
            try:
                rpc_config = await self.rpc.config()
            except APIError:
                pass

        if rpc_config is not None:
            try:
                return not rpc_config["CONFIG"][0]["RedLed"] == "off"
            except LookupError:
                pass

    async def _get_pools(self, rpc_pools: dict = None) -> List[PoolMetrics]:
        if rpc_pools is None:
            try:
                rpc_pools = await self.rpc.pools()
            except APIError:
                pass

        pools_data = []
        if rpc_pools is not None:
            try:
                pools = rpc_pools.get("POOLS", [])
                for pool_info in pools:
                    url = pool_info.get("URL")
                    pool_url = PoolUrl.from_str(url) if url else None
                    pool_data = PoolMetrics(
                        accepted=pool_info.get("Accepted"),
                        rejected=pool_info.get("Rejected"),
                        get_failures=pool_info.get("Get Failures"),
                        remote_failures=pool_info.get("Remote Failures"),
                        active=pool_info.get("Stratum Active"),
                        alive=pool_info.get("Status") == "Alive",
                        url=pool_url,
                        user=pool_info.get("User"),
                        index=pool_info.get("POOL"),
                    )
                    pools_data.append(pool_data)
            except LookupError:
                pass
        return pools_data



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M32V10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M32V10

    expected_chips = 78
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M32V20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M32V20

    expected_chips = 74
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M60VK10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60VK10

    expected_chips = 164
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M60VK20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60VK20

    expected_chips = 172
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M60VK30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60VK30

    expected_chips = 215
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M60VK40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60VK40

    expected_chips = 180
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M60VK6A(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60VK6A

    expected_chips = 172
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M60VL10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60VL10

    expected_chips = 111
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M60VL20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60VL20

    expected_chips = 117
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M60VL30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60VL30

    expected_chips = 123
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M60VL40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60VL40

    expected_chips = 129
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M60VL50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M60VL50

    expected_chips = 135
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M56SPlusPlusVK10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M56SPlusPlusVK10

    expected_chips = 160
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M56SPlusPlusVK30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M56SPlusPlusVK30

    expected_chips = 176
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M56SPlusPlusVK40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M56SPlusPlusVK40

    expected_chips = 132
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M56SPlusPlusVK50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M56SPlusPlusVK50

    expected_chips = 152
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256



from __future__ import annotations

from abc import ABC, abstractmethod

from pydantic import BaseModel, field_serializer
from typing_extensions import Self

from .unit.base import AlgoHashRateUnitType, GenericUnit


class AlgoHashRateType(BaseModel, ABC):
    unit: AlgoHashRateUnitType
    rate: float

    @field_serializer("unit")
    def serialize_unit(self, unit: AlgoHashRateUnitType):
        return unit.model_dump()

    @abstractmethod
    def into(self, other: "AlgoHashRateUnitType"):
        pass

    def auto_unit(self):
        if 1 < self.rate // int(self.unit.H) < 1000:
            return self.into(self.unit.H)
        if 1 < self.rate // int(self.unit.MH) < 1000:
            return self.into(self.unit.MH)
        if 1 < self.rate // int(self.unit.GH) < 1000:
            return self.into(self.unit.GH)
        if 1 < self.rate // int(self.unit.TH) < 1000:
            return self.into(self.unit.TH)
        if 1 < self.rate // int(self.unit.PH) < 1000:
            return self.into(self.unit.PH)
        if 1 < self.rate // int(self.unit.EH) < 1000:
            return self.into(self.unit.EH)
        if 1 < self.rate // int(self.unit.ZH) < 1000:
            return self.into(self.unit.ZH)
        return self

    def __float__(self):
        return float(self.rate)

    def __int__(self):
        return int(self.rate)

    def __repr__(self):
        return f"{self.rate} {str(self.unit)}"

    def __round__(self, n: int = None):
        return round(self.rate, n)

    def __add__(self, other: Self | int | float) -> Self:
        if isinstance(other, AlgoHashRateType):
            return self.__class__(
                rate=self.rate + other.into(self.unit).rate, unit=self.unit
            )
        return self.__class__(rate=self.rate + other, unit=self.unit)

    def __sub__(self, other: Self | int | float) -> Self:
        if isinstance(other, AlgoHashRateType):
            return self.__class__(
                rate=self.rate - other.into(self.unit).rate, unit=self.unit
            )
        return self.__class__(rate=self.rate - other, unit=self.unit)

    def __truediv__(self, other: Self | int | float) -> Self:
        if isinstance(other, AlgoHashRateType):
            return self.__class__(
                rate=self.rate / other.into(self.unit).rate, unit=self.unit
            )
        return self.__class__(rate=self.rate / other, unit=self.unit)

    def __floordiv__(self, other: Self | int | float) -> Self:
        if isinstance(other, AlgoHashRateType):
            return self.__class__(
                rate=self.rate // other.into(self.unit).rate, unit=self.unit
            )
        return self.__class__(rate=self.rate // other, unit=self.unit)

    def __mul__(self, other: Self | int | float) -> Self:
        if isinstance(other, AlgoHashRateType):
            return self.__class__(
                rate=self.rate * other.into(self.unit).rate, unit=self.unit
            )
        return self.__class__(rate=self.rate * other, unit=self.unit)


class GenericHashrate(AlgoHashRateType):
    rate: float = 0
    unit: GenericUnit = GenericUnit.H

    def into(self, other: GenericUnit):
        return self.__class__(
            rate=self.rate / (other.value / self.unit.value), unit=other
        )



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M39V10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M39V10

    expected_chips = 50
    expected_fans = 0
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M39V20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M39V20

    expected_chips = 54
    expected_fans = 0
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M39V30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M39V30

    expected_chips = 68
    expected_fans = 0
    expected_hashboards = 3
    algo = MinerAlgo.SHA256



from pydantic import BaseModel, ConfigDict, field_serializer

from pyasic.device.algorithm import MinerAlgoType
from pyasic.device.firmware import MinerFirmware
from pyasic.device.makes import MinerMake
from pyasic.device.models import MinerModelType


class DeviceInfo(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    make: MinerMake | None = None
    model: MinerModelType | None = None
    firmware: MinerFirmware | None = None
    algo: type[MinerAlgoType] | None = None

    @field_serializer("make")
    def serialize_make(self, make: MinerMake, _info):
        return str(make)

    @field_serializer("model")
    def serialize_model(self, model: MinerModelType, _info):
        return str(model)

    @field_serializer("firmware")
    def serialize_firmware(self, firmware: MinerFirmware, _info):
        return str(firmware)

    @field_serializer("algo")
    def serialize_algo(self, algo: MinerAlgoType, _info):
        return str(algo)



from pyasic.miners.backends import M6X
from pyasic.miners.device.models import M64VL30


class BTMinerM64VL30(M6X, M64VL30):
    pass


from pyasic.miners.device.models import M64VL40


class BTMinerM64VL40(M6X, M64VL40):
    pass



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M53VH30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53VH30

    expected_chips = 128
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M53VH40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53VH40

    expected_chips = 174
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M53VH50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53VH50

    expected_chips = 162
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M53VK30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53VK30

    expected_chips = 100
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256


class M53VK60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M53VK60

    expected_chips = 100
    expected_fans = 0
    expected_hashboards = 4
    algo = MinerAlgo.SHA256



import asyncio
import datetime
import json

from pyasic import get_miner


async def main():
    ip = "10.7.11.176"
    try:

        # settings.update("default_antminer_web_password", "admin")

        miner = await get_miner(ip=ip)

        print(f"Miner: {miner}")


        # resume = await miner.resume_mining()
        # print(f"Resume mining: {resume}")

        # stop = await miner.stop_mining()
        # print(f"Stop mining: {stop}")

        errors = await miner.get_errors()

        dataMiner = await miner.get_data()
        Hashboards = await miner.get_hashboards()

        miningMode = await miner.is_mining()
        sleepMode = await miner.is_sleep()

        print(f"Hashboards: {Hashboards}")
        print(f"Miner : {dataMiner}")
        print(f"Errors: {errors}")
        print(f"Is mining: {miningMode}")
        print(f"Sleep mode:  {sleepMode}")

        # Перевод в сон
        # try:
        #     goSleep = await miner.stop_mining()
        #     print(f"GoSleep: {goSleep}")
        # except Exception as e:
        #     print(f"GoSleep Error: {e}")




        # cur_pwd = "root"
        # new_pwd = "admin"
        # settings.update("default_antminer_web_password", "admin")

        # update_pwd = await miner.update_pwd(cur_pwd=cur_pwd, new_pwd=new_pwd)
        #
        # if update_pwd:
        #     print("update_pwd is true")
        # else:
        #     print("update_pwd is false")

        # try:
        #     data = await miner.resume_mining()
        # except Exception as e:
        #     print(e)

        # print(data)

        # def custom_serializer(o):
        #     if isinstance(o, datetime.datetime):
        #         return o.isoformat()
        #     try:
        #         return o.__dict__
        #     except AttributeError:
        #         return str(o)
        #
        # print(json.dumps(data, default=custom_serializer, indent=4))
    except Exception as e:
        print(f"Error:: {e}")


if __name__ == "__main__":
    asyncio.run(main())



from typing import List, Optional

from pyasic import MinerConfig
from pyasic.config import MiningModeConfig
from pyasic.data import Fan, HashBoard
from pyasic.data.pools import PoolMetrics, PoolUrl
from pyasic.device.algorithm import AlgoHashRate
from pyasic.errors import APIError
from pyasic.miners.data import DataFunction, DataLocations, DataOptions, WebAPICommand
from pyasic.miners.device.firmware import MaraFirmware
from pyasic.misc import merge_dicts
from pyasic.rpc.marathon import MaraRPCAPI
from pyasic.web.marathon import MaraWebAPI

MARA_DATA_LOC = DataLocations(
    **{
        str(DataOptions.MAC): DataFunction(
            "_get_mac",
            [WebAPICommand("web_overview", "overview")],
        ),
        str(DataOptions.FW_VERSION): DataFunction(
            "_get_fw_ver",
            [WebAPICommand("web_overview", "overview")],
        ),
        str(DataOptions.HOSTNAME): DataFunction(
            "_get_hostname",
            [WebAPICommand("web_network_config", "network_config")],
        ),
        str(DataOptions.HASHRATE): DataFunction(
            "_get_hashrate",
            [WebAPICommand("web_brief", "brief")],
        ),
        str(DataOptions.EXPECTED_HASHRATE): DataFunction(
            "_get_expected_hashrate",
            [WebAPICommand("web_brief", "brief")],
        ),
        str(DataOptions.HASHBOARDS): DataFunction(
            "_get_hashboards",
            [WebAPICommand("web_hashboards", "hashboards")],
        ),
        str(DataOptions.WATTAGE): DataFunction(
            "_get_wattage",
            [WebAPICommand("web_brief", "brief")],
        ),
        str(DataOptions.WATTAGE_LIMIT): DataFunction(
            "_get_wattage_limit",
            [WebAPICommand("web_miner_config", "miner_config")],
        ),
        str(DataOptions.FANS): DataFunction(
            "_get_fans",
            [WebAPICommand("web_fans", "fans")],
        ),
        str(DataOptions.FAULT_LIGHT): DataFunction(
            "_get_fault_light",
            [WebAPICommand("web_locate_miner", "locate_miner")],
        ),
        str(DataOptions.IS_MINING): DataFunction(
            "_is_mining",
            [WebAPICommand("web_brief", "brief")],
        ),
        str(DataOptions.UPTIME): DataFunction(
            "_get_uptime",
            [WebAPICommand("web_brief", "brief")],
        ),
        str(DataOptions.POOLS): DataFunction(
            "_get_pools",
            [WebAPICommand("web_pools", "pools")],
        ),
    }
)


class MaraMiner(MaraFirmware):
    _rpc_cls = MaraRPCAPI
    rpc: MaraRPCAPI
    _web_cls = MaraWebAPI
    web: MaraWebAPI

    data_locations = MARA_DATA_LOC

    async def fault_light_off(self) -> bool:
        res = await self.web.set_locate_miner(blinking=False)
        return res.get("blinking") is False

    async def fault_light_on(self) -> bool:
        res = await self.web.set_locate_miner(blinking=True)
        return res.get("blinking") is True

    async def get_config(self) -> MinerConfig:
        data = await self.web.get_miner_config()
        if data:
            self.config = MinerConfig.from_mara(data)
        return self.config

    async def send_config(self, config: MinerConfig, user_suffix: str = None) -> None:
        data = await self.web.get_miner_config()
        cfg_data = config.as_mara(user_suffix=user_suffix)
        merged_cfg = merge_dicts(data, cfg_data)
        await self.web.set_miner_config(**merged_cfg)

    async def set_power_limit(self, wattage: int) -> bool:
        cfg = await self.get_config()
        cfg.mining_mode = MiningModeConfig.power_tuning(power=wattage)
        await self.send_config(cfg)
        return True

    async def stop_mining(self) -> bool:
        data = await self.web.get_miner_config()
        data["mode"]["work-mode-selector"] = "Sleep"
        await self.web.set_miner_config(**data)
        return True

    async def resume_mining(self) -> bool:
        data = await self.web.get_miner_config()
        data["mode"]["work-mode-selector"] = "Auto"
        await self.web.set_miner_config(**data)
        return True

    async def reboot(self) -> bool:
        await self.web.reboot()
        return True

    async def restart_backend(self) -> bool:
        await self.web.reload()
        return True

    async def _get_wattage(self, web_brief: dict = None) -> Optional[int]:
        if web_brief is None:
            try:
                web_brief = await self.web.brief()
            except APIError:
                pass

        if web_brief is not None:
            try:
                return round(web_brief["power_consumption_estimated"])
            except LookupError:
                pass

    async def _is_mining(self, web_brief: dict = None) -> Optional[bool]:
        if web_brief is None:
            try:
                web_brief = await self.web.brief()
            except APIError:
                pass

        if web_brief is not None:
            try:
                return web_brief["status"] == "Mining"
            except LookupError:
                pass

    async def _get_uptime(self, web_brief: dict = None) -> Optional[int]:
        if web_brief is None:
            try:
                web_brief = await self.web.brief()
            except APIError:
                pass

        if web_brief is not None:
            try:
                return web_brief["elapsed"]
            except LookupError:
                pass

    async def _get_hashboards(self, web_hashboards: dict = None) -> List[HashBoard]:
        if self.expected_hashboards is None:
            return []

        hashboards = [
            HashBoard(slot=i, expected_chips=self.expected_chips)
            for i in range(self.expected_hashboards)
        ]

        if web_hashboards is None:
            try:
                web_hashboards = await self.web.hashboards()
            except APIError:
                pass

        if web_hashboards is not None:
            try:
                for hb in web_hashboards["hashboards"]:
                    idx = hb["index"]
                    hashboards[idx].hashrate = self.algo.hashrate(
                        rate=float(hb["hashrate_average"]), unit=self.algo.unit.GH
                    ).into(self.algo.unit.default)
                    hashboards[idx].temp = round(
                        sum(hb["temperature_pcb"]) / len(hb["temperature_pcb"])
                    )
                    hashboards[idx].chip_temp = round(
                        sum(hb["temperature_chip"]) / len(hb["temperature_chip"])
                    )
                    hashboards[idx].chips = hb["asic_num"]
                    hashboards[idx].serial_number = hb["serial_number"]
                    hashboards[idx].missing = False
            except LookupError:
                pass
        return hashboards

    async def _get_mac(self, web_overview: dict = None) -> Optional[str]:
        if web_overview is None:
            try:
                web_overview = await self.web.overview()
            except APIError:
                pass

        if web_overview is not None:
            try:
                return web_overview["mac"].upper()
            except LookupError:
                pass

    async def _get_fw_ver(self, web_overview: dict = None) -> Optional[str]:
        if web_overview is None:
            try:
                web_overview = await self.web.overview()
            except APIError:
                pass

        if web_overview is not None:
            try:
                return web_overview["version_firmware"]
            except LookupError:
                pass

    async def _get_hostname(self, web_network_config: dict = None) -> Optional[str]:
        if web_network_config is None:
            try:
                web_network_config = await self.web.get_network_config()
            except APIError:
                pass

        if web_network_config is not None:
            try:
                return web_network_config["hostname"]
            except LookupError:
                pass

    async def _get_hashrate(self, web_brief: dict = None) -> Optional[AlgoHashRate]:
        if web_brief is None:
            try:
                web_brief = await self.web.brief()
            except APIError:
                pass

        if web_brief is not None:
            try:
                return self.algo.hashrate(
                    rate=float(web_brief["hashrate_realtime"]), unit=self.algo.unit.TH
                ).into(self.algo.unit.default)
            except LookupError:
                pass

    async def _get_fans(self, web_fans: dict = None) -> List[Fan]:
        if self.expected_fans is None:
            return []

        if web_fans is None:
            try:
                web_fans = await self.web.fans()
            except APIError:
                pass

        if web_fans is not None:
            fans = []
            for n in range(self.expected_fans):
                try:
                    fans.append(Fan(speed=web_fans["fans"][n]["current_speed"]))
                except (IndexError, KeyError):
                    pass
            return fans
        return [Fan() for _ in range(self.expected_fans)]

    async def _get_fault_light(self, web_locate_miner: dict = None) -> bool:
        if web_locate_miner is None:
            try:
                web_locate_miner = await self.web.get_locate_miner()
            except APIError:
                pass

        if web_locate_miner is not None:
            try:
                return web_locate_miner["blinking"]
            except LookupError:
                pass
        return False

    async def _get_expected_hashrate(
        self, web_brief: dict = None
    ) -> Optional[AlgoHashRate]:
        if web_brief is None:
            try:
                web_brief = await self.web.brief()
            except APIError:
                pass

        if web_brief is not None:
            try:
                return self.algo.hashrate(
                    rate=float(web_brief["hashrate_ideal"]), unit=self.algo.unit.GH
                ).into(self.algo.unit.default)
            except LookupError:
                pass

    async def _get_wattage_limit(
        self, web_miner_config: dict = None
    ) -> Optional[AlgoHashRate]:
        if web_miner_config is None:
            try:
                web_miner_config = await self.web.get_miner_config()
            except APIError:
                pass

        if web_miner_config is not None:
            try:
                return web_miner_config["mode"]["concorde"]["power-target"]
            except LookupError:
                pass

    async def _get_pools(self, web_pools: list = None) -> List[PoolMetrics]:
        if web_pools is None:
            try:
                web_pools = await self.web.pools()
            except APIError:
                return []

        active_pool_index = None
        highest_priority = float("inf")

        for pool_info in web_pools:
            if (
                pool_info.get("status") == "Alive"
                and pool_info.get("priority", float("inf")) < highest_priority
            ):
                highest_priority = pool_info["priority"]
                active_pool_index = pool_info["index"]

        pools_data = []
        if web_pools is not None:
            try:
                for pool_info in web_pools:
                    url = pool_info.get("url")
                    pool_url = PoolUrl.from_str(url) if url else None
                    pool_data = PoolMetrics(
                        accepted=pool_info.get("accepted"),
                        rejected=pool_info.get("rejected"),
                        get_failures=pool_info.get("stale"),
                        remote_failures=pool_info.get("discarded"),
                        active=pool_info.get("index") == active_pool_index,
                        alive=pool_info.get("status") == "Alive",
                        url=pool_url,
                        user=pool_info.get("user"),
                        index=pool_info.get("index"),
                    )
                    pools_data.append(pool_data)
            except LookupError:
                pass
        return pools_data



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------
from __future__ import annotations

from pyasic.config.base import MinerConfigValue


class ScalingShutdown(MinerConfigValue):
    enabled: bool = False
    duration: int | None = None

    @classmethod
    def from_dict(cls, dict_conf: dict | None) -> "ScalingShutdown":
        return cls(
            enabled=dict_conf.get("enabled", False), duration=dict_conf.get("duration")
        )

    @classmethod
    def from_bosminer(cls, power_scaling_conf: dict):
        sd_enabled = power_scaling_conf.get("shutdown_enabled")
        if sd_enabled is not None:
            return cls(
                enabled=sd_enabled, duration=power_scaling_conf.get("shutdown_duration")
            )
        return None

    @classmethod
    def from_boser(cls, power_scaling_conf: dict):
        sd_enabled = power_scaling_conf.get("shutdownEnabled")
        if sd_enabled is not None:
            try:
                return cls(
                    enabled=sd_enabled,
                    duration=power_scaling_conf["shutdownDuration"]["hours"],
                )
            except KeyError:
                return cls(enabled=sd_enabled)
        return None

    def as_bosminer(self) -> dict:
        cfg = {"shutdown_enabled": self.enabled}

        if self.duration is not None:
            cfg["shutdown_duration"] = self.duration

        return cfg

    def as_boser(self) -> dict:
        return {"enable_shutdown": self.enabled, "shutdown_duration": self.duration}


class ScalingConfig(MinerConfigValue):
    step: int | None = None
    minimum: int | None = None
    shutdown: ScalingShutdown | None = None

    @classmethod
    def from_dict(cls, dict_conf: dict | None) -> "ScalingConfig":
        cls_conf = {
            "step": dict_conf.get("step"),
            "minimum": dict_conf.get("minimum"),
        }
        shutdown = dict_conf.get("shutdown")
        if shutdown is not None:
            cls_conf["shutdown"] = ScalingShutdown.from_dict(shutdown)
        return cls(**cls_conf)

    @classmethod
    def from_bosminer(cls, toml_conf: dict, mode: str | None = None):
        if mode == "power":
            return cls._from_bosminer_power(toml_conf)
        if mode == "hashrate":
            # not implemented yet
            pass

    @classmethod
    def _from_bosminer_power(cls, toml_conf: dict):
        power_scaling = toml_conf.get("power_scaling")
        if power_scaling is None:
            power_scaling = toml_conf.get("performance_scaling")
        if power_scaling is not None:
            enabled = power_scaling.get("enabled")
            if not enabled:
                return None
            power_step = power_scaling.get("power_step")
            min_power = power_scaling.get("min_psu_power_limit")
            if min_power is None:
                min_power = power_scaling.get("min_power_target")
            sd_mode = ScalingShutdown.from_bosminer(power_scaling)

            return cls(step=power_step, minimum=min_power, shutdown=sd_mode)

    @classmethod
    def from_boser(cls, grpc_miner_conf: dict, mode: str | None = None):
        if mode == "power":
            return cls._from_boser_power(grpc_miner_conf)
        if mode == "hashrate":
            # not implemented yet
            pass

    @classmethod
    def _from_boser_power(cls, grpc_miner_conf: dict):
        try:
            dps_conf = grpc_miner_conf["dps"]
            if not dps_conf.get("enabled", False):
                return None
        except LookupError:
            return None

        conf = {"shutdown": ScalingShutdown.from_boser(dps_conf)}

        if dps_conf.get("minPowerTarget") is not None:
            conf["minimum"] = dps_conf["minPowerTarget"]["watt"]
        if dps_conf.get("powerStep") is not None:
            conf["step"] = dps_conf["powerStep"]["watt"]
        return cls(**conf)



from pyasic.miners.backends import M6X
from pyasic.miners.device.models import M66SVK20


class BTMinerM66SVK20(M6X, M66SVK20):
    pass


from pyasic.miners.device.models import M66SVK30


class BTMinerM66SVK30(M6X, M66SVK30):
    pass


from pyasic.miners.device.models import M66SVK40


class BTMinerM66SVK40(M6X, M66SVK40):
    pass


from pyasic.miners.device.models import M66SVK50


class BTMinerM66SVK50(M6X, M66SVK50):
    pass


from pyasic.miners.device.models import M66SVK60


class BTMinerM66SVK60(M6X, M66SVK60):
    pass


from pyasic.miners.device.models import M66SVL10


class BTMinerM66SVL10(M6X, M66SVL10):
    pass


from pyasic.miners.device.models import M66SVL20


class BTMinerM66SVL20(M6X, M66SVL20):
    pass


from pyasic.miners.device.models import M66SVL30


class BTMinerM66SVL30(M6X, M66SVL30):
    pass


from pyasic.miners.device.models import M66SVL40


class BTMinerM66SVL40(M6X, M66SVL40):
    pass


from pyasic.miners.device.models import M66SVL50


class BTMinerM66SVL50(M6X, M66SVL50):
    pass



from pyasic.miners.backends import M6X
from pyasic.miners.device.models import M60VK10


class BTMinerM60VK10(M6X, M60VK10):
    pass


from pyasic.miners.device.models import M60VK20


class BTMinerM60VK20(M6X, M60VK20):
    pass


from pyasic.miners.device.models import M60VK30


class BTMinerM60VK30(M6X, M60VK30):
    pass


from pyasic.miners.device.models import M60VK40


class BTMinerM60VK40(M6X, M60VK40):
    pass


from pyasic.miners.device.models import M60VK6A


class BTMinerM60VK6A(M6X, M60VK6A):
    pass


from pyasic.miners.device.models import M60VL10


class BTMinerM60VL10(M6X, M60VL10):
    pass


from pyasic.miners.device.models import M60VL20


class BTMinerM60VL20(M6X, M60VL20):
    pass


from pyasic.miners.device.models import M60VL30


class BTMinerM60VL30(M6X, M60VL30):
    pass


from pyasic.miners.device.models import M60VL40


class BTMinerM60VL40(M6X, M60VL40):
    pass


from pyasic.miners.device.models import M60VL50


class BTMinerM60VL50(M6X, M60VL50):
    pass



from pyasic.device.algorithm import MinerAlgo
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M30SPlusV10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV10

    expected_chips = 215
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusV100(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV100

    expected_chips = 215
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusV20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV20

    expected_chips = 255
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusV30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV30

    expected_chips = None
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusV40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV40

    expected_chips = 235
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusV50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV50

    expected_chips = 225
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusV60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV60

    expected_chips = 245
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusV70(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV70

    expected_chips = 235
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusV80(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV80

    expected_chips = 245
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusV90(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV90

    expected_chips = 225
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVE100(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVE100

    expected_chips = None
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVE30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVE30

    expected_chips = 148
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVE40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVE40

    expected_chips = 156
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVE50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVE50

    expected_chips = 164
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVE60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVE60

    expected_chips = 172
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVE70(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVE70

    expected_chips = None
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVE80(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVE80

    expected_chips = None
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVE90(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVE90

    expected_chips = None
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVF20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVF20

    expected_chips = 111
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVF30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVF30

    expected_chips = 117
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVG20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVG20

    expected_chips = 82
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVG30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVG30

    expected_chips = 78
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVG40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVG40

    expected_chips = 105
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVG50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVG50

    expected_chips = 111
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVG60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVG60

    expected_chips = 86
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVH10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVH10

    expected_chips = 64
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVH20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVH20

    expected_chips = 66
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVH30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVH30

    expected_chips = 70
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVH40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVH40

    expected_chips = 74
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVH50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVH50

    expected_chips = 64
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVH60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVH60

    expected_chips = 66
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVH70(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVH70

    expected_chips = 70
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVI30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVI30

    expected_chips = 86
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVJ30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVJ30

    expected_chips = 105
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256


class M30SPlusVJ40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVJ40

    expected_chips = 117
    expected_fans = 2
    expected_hashboards = 3
    algo = MinerAlgo.SHA256



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------

import logging
from pathlib import Path
from typing import List, Optional

import aiofiles

from pyasic.config import MinerConfig, MiningModeConfig
from pyasic.data import Fan, HashBoard
from pyasic.data.error_codes import MinerErrorData, WhatsminerError
from pyasic.data.pools import PoolMetrics, PoolUrl
from pyasic.device.algorithm import AlgoHashRate
from pyasic.errors import APIError
from pyasic.miners.data import DataFunction, DataLocations, DataOptions, RPCAPICommand
from pyasic.miners.device.firmware import StockFirmware
from pyasic.rpc.btminer import BTMinerRPCAPI

BTMINER_DATA_LOC = DataLocations(
    **{
        str(DataOptions.MAC): DataFunction(
            "_get_mac",
            [
                RPCAPICommand("rpc_summary", "summary"),
                RPCAPICommand("rpc_get_miner_info", "get_miner_info"),
            ],
        ),
        str(DataOptions.API_VERSION): DataFunction(
            "_get_api_ver",
            [RPCAPICommand("rpc_get_version", "get_version")],
        ),
        str(DataOptions.FW_VERSION): DataFunction(
            "_get_fw_ver",
            [
                RPCAPICommand("rpc_get_version", "get_version"),
                RPCAPICommand("rpc_summary", "summary"),
            ],
        ),
        str(DataOptions.HOSTNAME): DataFunction(
            "_get_hostname",
            [RPCAPICommand("rpc_get_miner_info", "get_miner_info")],
        ),
        str(DataOptions.HASHRATE): DataFunction(
            "_get_hashrate",
            [RPCAPICommand("rpc_summary", "summary")],
        ),
        str(DataOptions.EXPECTED_HASHRATE): DataFunction(
            "_get_expected_hashrate",
            [RPCAPICommand("rpc_summary", "summary")],
        ),
        str(DataOptions.HASHBOARDS): DataFunction(
            "_get_hashboards",
            [RPCAPICommand("rpc_devs", "devs")],
        ),
        str(DataOptions.ENVIRONMENT_TEMP): DataFunction(
            "_get_env_temp",
            [RPCAPICommand("rpc_summary", "summary")],
        ),
        str(DataOptions.WATTAGE): DataFunction(
            "_get_wattage",
            [RPCAPICommand("rpc_summary", "summary")],
        ),
        str(DataOptions.WATTAGE_LIMIT): DataFunction(
            "_get_wattage_limit",
            [RPCAPICommand("rpc_summary", "summary")],
        ),
        str(DataOptions.FANS): DataFunction(
            "_get_fans",
            [
                RPCAPICommand("rpc_summary", "summary"),
                RPCAPICommand("rpc_get_psu", "get_psu"),
            ],
        ),
        str(DataOptions.FAN_PSU): DataFunction(
            "_get_fan_psu",
            [
                RPCAPICommand("rpc_summary", "summary"),
                RPCAPICommand("rpc_get_psu", "get_psu"),
            ],
        ),
        str(DataOptions.ERRORS): DataFunction(
            "_get_errors",
            [
                RPCAPICommand("rpc_get_error_code", "get_error_code"),
                RPCAPICommand("rpc_summary", "summary"),
            ],
        ),
        str(DataOptions.FAULT_LIGHT): DataFunction(
            "_get_fault_light",
            [RPCAPICommand("rpc_get_miner_info", "get_miner_info")],
        ),
        str(DataOptions.IS_MINING): DataFunction(
            "_is_mining",
            [RPCAPICommand("rpc_status", "status")],
        ),
        str(DataOptions.IS_SLEEP): DataFunction(
            "_is_sleep",
            [],
        ),
        str(DataOptions.UPTIME): DataFunction(
            "_get_uptime",
            [RPCAPICommand("rpc_summary", "summary")],
        ),
        str(DataOptions.POOLS): DataFunction(
            "_get_pools",
            [RPCAPICommand("rpc_pools", "pools")],
        ),
    }
)


class BTMiner(StockFirmware):
    """Base handler for BTMiner based miners."""

    _rpc_cls = BTMinerRPCAPI
    rpc: BTMinerRPCAPI

    data_locations = BTMINER_DATA_LOC

    supports_shutdown = True
    supports_power_modes = True

    async def _reset_rpc_pwd_to_admin(self, pwd: str):
        try:
            data = await self.rpc.update_pwd(pwd, "admin")
        except APIError:
            return False
        if data:
            if "Code" in data.keys():
                if data["Code"] == 131:
                    return True
        return False

    async def fault_light_off(self) -> bool:
        try:
            data = await self.rpc.set_led(auto=True)
        except APIError:
            return False
        if data:
            if "Code" in data.keys():
                if data["Code"] == 131:
                    self.light = False
                    return True
        return False

    async def fault_light_on(self) -> bool:
        try:
            data = await self.rpc.set_led(auto=False)
            await self.rpc.set_led(
                auto=False, color="green", start=0, period=1, duration=0
            )
        except APIError:
            return False
        if data:
            if "Code" in data.keys():
                if data["Code"] == 131:
                    self.light = True
                    return True
        return False

    async def reboot(self) -> bool:
        try:
            data = await self.rpc.reboot()
        except APIError:
            return False
        if data.get("Msg"):
            if data["Msg"] == "API command OK":
                return True
        return False

    async def restart_backend(self) -> bool:
        try:
            data = await self.rpc.restart()
        except APIError:
            return False
        if data.get("Msg"):
            if data["Msg"] == "API command OK":
                return True
        return False

    async def stop_mining(self) -> bool:
        try:
            data = await self.rpc.power_off(respbefore=True)
        except APIError:
            return False
        if data.get("Msg"):
            if data["Msg"] == "API command OK":
                return True
        return False

    async def resume_mining(self) -> bool:
        try:
            data = await self.rpc.power_on()
        except APIError:
            return False
        if data.get("Msg"):
            if data["Msg"] == "API command OK":
                return True
        return False

    async def send_config(self, config: MinerConfig, user_suffix: str = None) -> None:
        self.config = config

        conf = config.as_wm(user_suffix=user_suffix)
        pools_conf = conf["pools"]

        try:
            await self.rpc.update_pools(**pools_conf)

            if conf["mode"] == "normal":
                await self.rpc.set_normal_power()
            elif conf["mode"] == "high":
                await self.rpc.set_high_power()
            elif conf["mode"] == "low":
                await self.rpc.set_low_power()
            elif conf["mode"] == "power_tuning":
                await self.rpc.adjust_power_limit(conf["power_tuning"]["wattage"])
        except APIError:
            # cannot update, no API access usually
            pass

    async def get_config(self) -> MinerConfig:
        pools = None
        summary = None
        status = None
        try:
            data = await self.rpc.multicommand("pools", "summary", "status")
            pools = data["pools"][0]
            summary = data["summary"][0]
            status = data["status"][0]
        except APIError as e:
            logging.warning(e)
        except LookupError:
            pass

        if pools is not None:
            cfg = MinerConfig.from_api(pools)
        else:
            cfg = MinerConfig()

        is_mining = await self._is_mining(status)
        if not is_mining:
            cfg.mining_mode = MiningModeConfig.sleep()
            return cfg

        if summary is not None:
            mining_mode = None
            try:
                mining_mode = summary["SUMMARY"][0]["Power Mode"]
            except LookupError:
                pass

            if mining_mode == "High":
                cfg.mining_mode = MiningModeConfig.high()
                return cfg
            elif mining_mode == "Low":
                cfg.mining_mode = MiningModeConfig.low()
                return cfg
            try:
                power_lim = summary["SUMMARY"][0]["Power Limit"]
            except LookupError:
                power_lim = None

            if power_lim is None:
                cfg.mining_mode = MiningModeConfig.normal()
                return cfg

            cfg.mining_mode = MiningModeConfig.power_tuning(power=power_lim)
            self.config = cfg
            return self.config

    async def set_power_limit(self, wattage: int) -> bool:
        try:
            await self.rpc.adjust_power_limit(wattage)
        except Exception as e:
            logging.warning(f"{self} set_power_limit: {e}")
            return False
        else:
            return True

    ##################################################
    ### DATA GATHERING FUNCTIONS (get_{some_data}) ###
    ##################################################

    async def _get_mac(
        self, rpc_summary: dict = None, rpc_get_miner_info: dict = None
    ) -> Optional[str]:
        if rpc_get_miner_info is None:
            try:
                rpc_get_miner_info = await self.rpc.get_miner_info()
            except APIError:
                pass

        if rpc_get_miner_info is not None:
            try:
                mac = rpc_get_miner_info["Msg"]["mac"]
                return str(mac).upper()
            except KeyError:
                pass

        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        if rpc_summary is not None:
            try:
                mac = rpc_summary["SUMMARY"][0]["MAC"]
                return str(mac).upper()
            except LookupError:
                pass

    async def _get_api_ver(self, rpc_get_version: dict = None) -> Optional[str]:
        if rpc_get_version is None:
            try:
                rpc_get_version = await self.rpc.get_version()
            except APIError:
                pass

        if rpc_get_version is not None:
            if "Code" in rpc_get_version.keys():
                if rpc_get_version["Code"] == 131:
                    try:
                        rpc_ver = rpc_get_version["Msg"]
                        if not isinstance(rpc_ver, str):
                            rpc_ver = rpc_ver["rpc_ver"]
                        self.api_ver = rpc_ver.replace("whatsminer v", "")
                    except (KeyError, TypeError):
                        pass
                    else:
                        self.rpc.rpc_ver = self.api_ver
                        return self.api_ver

        return self.api_ver

    async def _get_fw_ver(
        self, rpc_get_version: dict = None, rpc_summary: dict = None
    ) -> Optional[str]:
        if rpc_get_version is None:
            try:
                rpc_get_version = await self.rpc.get_version()
            except APIError:
                pass

        if rpc_get_version is not None:
            if "Code" in rpc_get_version.keys():
                if rpc_get_version["Code"] == 131:
                    try:
                        self.fw_ver = rpc_get_version["Msg"]["fw_ver"]
                    except (KeyError, TypeError):
                        pass
                    else:
                        return self.fw_ver

        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        if rpc_summary:
            try:
                self.fw_ver = rpc_summary["SUMMARY"][0]["Firmware Version"].replace(
                    "'", ""
                )
            except LookupError:
                pass

        return self.fw_ver

    async def _get_hostname(self, rpc_get_miner_info: dict = None) -> Optional[str]:
        hostname = None
        if rpc_get_miner_info is None:
            try:
                rpc_get_miner_info = await self.rpc.get_miner_info()
            except APIError:
                return None  # only one way to get this

        if rpc_get_miner_info is not None:
            try:
                hostname = rpc_get_miner_info["Msg"]["hostname"]
            except KeyError:
                return None

        return hostname

    async def _get_hashrate(self, rpc_summary: dict = None) -> Optional[AlgoHashRate]:
        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        if rpc_summary is not None:
            try:
                return self.algo.hashrate(
                    rate=float(rpc_summary["SUMMARY"][0]["MHS 1m"]),
                    unit=self.algo.unit.MH,
                ).into(self.algo.unit.default)
            except LookupError:
                pass

    async def _get_hashboards(self, rpc_devs: dict = None) -> List[HashBoard]:
        if self.expected_hashboards is None:
            return []

        hashboards = [
            HashBoard(slot=i, expected_chips=self.expected_chips)
            for i in range(self.expected_hashboards)
        ]

        if rpc_devs is None:
            try:
                rpc_devs = await self.rpc.devs()
            except APIError:
                pass

        if rpc_devs is not None:
            try:
                for board in rpc_devs["DEVS"]:
                    asc = board.get("ASC")
                    if asc is None:
                        asc = board["Slot"]
                    if len(hashboards) < asc + 1:
                        hashboards.append(
                            HashBoard(slot=asc, expected_chips=self.expected_chips)
                        )
                        self.expected_hashboards += 1
                    hashboards[asc].chip_temp = round(board["Chip Temp Avg"])
                    hashboards[asc].temp = round(board["Temperature"])
                    hashboards[asc].hashrate = self.algo.hashrate(
                        rate=float(board["MHS 1m"]), unit=self.algo.unit.MH
                    ).into(self.algo.unit.default)
                    hashboards[asc].chips = board["Effective Chips"]
                    hashboards[asc].serial_number = board["PCB SN"]
                    hashboards[asc].missing = False
            except LookupError:
                pass

        return hashboards

    async def _get_env_temp(self, rpc_summary: dict = None) -> Optional[float]:
        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        if rpc_summary is not None:
            try:
                return rpc_summary["SUMMARY"][0]["Env Temp"]
            except LookupError:
                pass

    async def _get_wattage(self, rpc_summary: dict = None) -> Optional[int]:
        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        if rpc_summary is not None:
            try:
                wattage = rpc_summary["SUMMARY"][0]["Power"]
                return wattage if not wattage == -1 else None
            except LookupError:
                pass

    async def _get_wattage_limit(self, rpc_summary: dict = None) -> Optional[int]:
        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        if rpc_summary is not None:
            try:
                return rpc_summary["SUMMARY"][0]["Power Limit"]
            except LookupError:
                pass

    async def _get_fans(
        self, rpc_summary: dict = None, rpc_get_psu: dict = None
    ) -> List[Fan]:
        if self.expected_fans is None:
            return []

        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        fans = [Fan() for _ in range(self.expected_fans)]
        if rpc_summary is not None:
            try:
                if self.expected_fans > 0:
                    fans = [
                        Fan(speed=rpc_summary["SUMMARY"][0].get("Fan Speed In", 0)),
                        Fan(speed=rpc_summary["SUMMARY"][0].get("Fan Speed Out", 0)),
                    ]
            except LookupError:
                pass

        return fans

    async def _get_fan_psu(
        self, rpc_summary: dict = None, rpc_get_psu: dict = None
    ) -> Optional[int]:
        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        if rpc_summary is not None:
            try:
                return int(rpc_summary["SUMMARY"][0]["Power Fanspeed"])
            except LookupError:
                pass

        if rpc_get_psu is None:
            try:
                rpc_get_psu = await self.rpc.get_psu()
            except APIError:
                pass

        if rpc_get_psu is not None:
            try:
                return int(rpc_get_psu["Msg"]["fan_speed"])
            except (KeyError, TypeError):
                pass

    async def _get_errors(
        self, rpc_summary: dict = None, rpc_get_error_code: dict = None
    ) -> List[MinerErrorData]:
        errors = []
        if rpc_get_error_code is None and rpc_summary is None:
            try:
                rpc_get_error_code = await self.rpc.get_error_code()
            except APIError:
                pass

        if rpc_get_error_code is not None:
            try:
                for err in rpc_get_error_code["Msg"]["error_code"]:
                    if isinstance(err, dict):
                        for code in err:
                            errors.append(WhatsminerError(error_code=int(code)))
                    else:
                        errors.append(WhatsminerError(error_code=int(err)))
            except KeyError:
                pass

        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        if rpc_summary is not None:
            try:
                for i in range(rpc_summary["SUMMARY"][0]["Error Code Count"]):
                    err = rpc_summary["SUMMARY"][0].get(f"Error Code {i}")
                    if err:
                        errors.append(WhatsminerError(error_code=err))
            except (LookupError, ValueError, TypeError):
                pass
        return errors

    async def _get_expected_hashrate(
        self, rpc_summary: dict = None
    ) -> Optional[AlgoHashRate]:
        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        if rpc_summary is not None:
            try:
                expected_hashrate = rpc_summary["SUMMARY"][0]["Factory GHS"]
                if expected_hashrate:
                    return self.algo.hashrate(
                        rate=float(expected_hashrate), unit=self.algo.unit.GH
                    ).into(self.algo.unit.default)

            except LookupError:
                pass

    async def _get_fault_light(self, rpc_get_miner_info: dict = None) -> Optional[bool]:
        if rpc_get_miner_info is None:
            try:
                rpc_get_miner_info = await self.rpc.get_miner_info()
            except APIError:
                if not self.light:
                    self.light = False

        if rpc_get_miner_info is not None:
            try:
                self.light = not (rpc_get_miner_info["Msg"]["ledstat"] == "auto")
            except KeyError:
                pass

        return self.light if self.light else False

    async def set_static_ip(
        self,
        ip: str,
        dns: str,
        gateway: str,
        subnet_mask: str = "255.255.255.0",
        hostname: str = None,
    ):
        if not hostname:
            hostname = await self.get_hostname()
        await self.rpc.net_config(
            ip=ip, mask=subnet_mask, dns=dns, gate=gateway, host=hostname, dhcp=False
        )

    async def set_dhcp(self, hostname: str = None):
        if hostname:
            await self.set_hostname(hostname)
        await self.rpc.net_config()

    async def set_hostname(self, hostname: str):
        await self.rpc.set_hostname(hostname)

    async def _is_mining(self, rpc_status: dict = None) -> Optional[bool]:
        if rpc_status is None:
            try:
                rpc_status = await self.rpc.status()
            except APIError:
                pass

        if rpc_status is not None:
            try:
                if rpc_status["Msg"].get("btmineroff"):
                    try:
                        await self.rpc.devdetails()
                    except APIError:
                        return False
                    return True
                return True if rpc_status["Msg"]["mineroff"] == "false" else False
            except LookupError:
                pass

    async def _is_sleep(self, rpc_status: dict = None) -> bool:
        # 1) Если нет переданных данных, запрашиваем статус у API
        if rpc_status is None:
            try:
                rpc_status = await self.rpc.status()
            except APIError:
                # при ошибке запроса считаем, что спящего режима нет
                return False

        # 2) Пытаемся достать btmineroff
        try:
            btmineroff = rpc_status.get("Msg", {}).get("btmineroff")
        except AttributeError:
            # rpc_status не тот формат — возвращаем False
            return False

        # 3) Обрабатываем разные типы и значения
        if isinstance(btmineroff, bool):
            # если вдруг API вернул уже булево
            return btmineroff

        if isinstance(btmineroff, str):
            val = btmineroff.strip().lower()
            if val == "true":
                return True
            if val == "false":
                return False

        # 4) Во всех остальных случаях (None, число, неожиданный текст) — считаем False
        return False

    async def _get_uptime(self, rpc_summary: dict = None) -> Optional[int]:
        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        if rpc_summary is not None:
            try:
                return int(rpc_summary["SUMMARY"][0]["Elapsed"])
            except LookupError:
                pass

    async def _get_pools(self, rpc_pools: dict = None) -> List[PoolMetrics]:
        if rpc_pools is None:
            try:
                rpc_pools = await self.rpc.pools()
            except APIError:
                pass

        pools_data = []
        if rpc_pools is not None:
            try:
                pools = rpc_pools.get("POOLS", [])
                for pool_info in pools:
                    url = pool_info.get("URL")
                    pool_url = PoolUrl.from_str(url) if url else None
                    pool_data = PoolMetrics(
                        accepted=pool_info.get("Accepted"),
                        rejected=pool_info.get("Rejected"),
                        get_failures=pool_info.get("Get Failures"),
                        remote_failures=pool_info.get("Remote Failures"),
                        active=pool_info.get("Stratum Active"),
                        alive=pool_info.get("Status") == "Alive",
                        url=pool_url,
                        user=pool_info.get("User"),
                        index=pool_info.get("POOL"),
                    )
                    pools_data.append(pool_data)
            except LookupError:
                pass
        return pools_data

    async def upgrade_firmware(self, file: Path) -> str:
        """
        Upgrade the firmware of the Whatsminer device.

        Args:
            file (Path): The local file path of the firmware to be uploaded.

        Returns:
            str: Confirmation message after upgrading the firmware.
        """
        try:
            logging.info("Starting firmware upgrade process for Whatsminer.")

            if not file:
                raise ValueError("File location must be provided for firmware upgrade.")

            # Read the firmware file contents
            async with aiofiles.open(file, "rb") as f:
                upgrade_contents = await f.read()

            result = await self.rpc.update_firmware(upgrade_contents)

            logging.info(
                "Firmware upgrade process completed successfully for Whatsminer."
            )
            return result
        except FileNotFoundError as e:
            logging.error(f"File not found during the firmware upgrade process: {e}")
            raise
        except ValueError as e:
            logging.error(
                f"Validation error occurred during the firmware upgrade process: {e}"
            )
            raise
        except OSError as e:
            logging.error(f"OS error occurred during the firmware upgrade process: {e}")
            raise
        except Exception as e:
            logging.error(
                f"An unexpected error occurred during the firmware upgrade process: {e}",
                exc_info=True,
            )
            raise



from pyasic.miners.backends import M6X
from pyasic.miners.device.models import M66VK20


class BTMinerM66VK20(M6X, M66VK20):
    pass


from pyasic.miners.device.models import M66VK30


class BTMinerM66VK30(M6X, M66VK30):
    pass


from pyasic.miners.device.models import M66VL20


class BTMinerM66VL20(M6X, M66VL20):
    pass


from pyasic.miners.device.models import M66VL30


class BTMinerM66VL30(M6X, M66VL30):
    pass



from pyasic.miners.backends import M5X
from pyasic.miners.device.models import M56SVH30


class BTMinerM56SVH30(M5X, M56SVH30):
    pass


from pyasic.miners.device.models import M56SVJ30


class BTMinerM56SVJ30(M5X, M56SVJ30):
    pass


from pyasic.miners.device.models import M56SVJ40


class BTMinerM56SVJ40(M5X, M56SVJ40):
    pass



from pyasic.miners.backends import M6X
from pyasic.miners.device.models import M63SPlusVK30


class BTMinerM63SPlusVK30(M6X, M63SPlusVK30):
    pass


from pyasic.miners.device.models import M63SPlusVL10


class BTMinerM63SPlusVL10(M6X, M63SPlusVL10):
    pass


from pyasic.miners.device.models import M63SPlusVL20


class BTMinerM63SPlusVL20(M6X, M63SPlusVL20):
    pass


from pyasic.miners.device.models import M63SPlusVL30


class BTMinerM63SPlusVL30(M6X, M63SPlusVL30):
    pass


from pyasic.miners.device.models import M63SPlusVL50


class BTMinerM63SPlusVL50(M6X, M63SPlusVL50):
    pass



from pyasic.miners.backends import M5X
from pyasic.miners.device.models import M50SPlusVH30


class BTMinerM50SPlusVH30(M5X, M50SPlusVH30):
    pass


from pyasic.miners.device.models import M50SPlusVH40


class BTMinerM50SPlusVH40(M5X, M50SPlusVH40):
    pass


from pyasic.miners.device.models import M50SPlusVJ30


class BTMinerM50SPlusVJ30(M5X, M50SPlusVJ30):
    pass


from pyasic.miners.device.models import M50SPlusVJ40


class BTMinerM50SPlusVJ40(M5X, M50SPlusVJ40):
    pass


from pyasic.miners.device.models import M50SPlusVJ60


class BTMinerM50SPlusVJ60(M5X, M50SPlusVJ60):
    pass


from pyasic.miners.device.models import M50SPlusVK10


class BTMinerM50SPlusVK10(M5X, M50SPlusVK10):
    pass


from pyasic.miners.device.models import M50SPlusVK20


class BTMinerM50SPlusVK20(M5X, M50SPlusVK20):
    pass


from pyasic.miners.device.models import M50SPlusVK30


class BTMinerM50SPlusVK30(M5X, M50SPlusVK30):
    pass


from pyasic.miners.device.models import M50SPlusVL10


class BTMinerM50SPlusVL10(M5X, M50SPlusVL10):
    pass


from pyasic.miners.device.models import M50SPlusVL20


class BTMinerM50SPlusVL20(M5X, M50SPlusVL20):
    pass


from pyasic.miners.device.models import M50SPlusVL30


class BTMinerM50SPlusVL30(M5X, M50SPlusVL30):
    pass
